# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Union

from supertokens_python.normalised_url_path import NormalisedURLPath

if TYPE_CHECKING:
    from supertokens_python.querier import Querier

from .interfaces import RecipeInterface, SignInUpOkResult
from .types import ThirdPartyInfo, User


class RecipeImplementation(RecipeInterface):
    def __init__(self, querier: Querier):
        super().__init__()
        self.querier = querier

    async def get_user_by_id(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        params = {"userId": user_id}
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/user"), params
        )
        if "status" in response and response["status"] == "OK":
            return User(
                response["user"]["id"],
                response["user"]["email"],
                response["user"]["timeJoined"],
                ThirdPartyInfo(
                    response["user"]["thirdParty"]["userId"],
                    response["user"]["thirdParty"]["id"],
                ),
            )
        return None

    async def get_users_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> List[User]:
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/users/by-email"), {"email": email}
        )
        users: List[User] = []
        users_list: List[Dict[str, Any]] = (
            response["users"] if "users" in response else []
        )
        for user in users_list:
            users.append(
                User(
                    user["id"],
                    user["email"],
                    user["timeJoined"],
                    ThirdPartyInfo(
                        user["thirdParty"]["userId"], user["thirdParty"]["id"]
                    ),
                )
            )
        return users

    async def get_user_by_thirdparty_info(
        self,
        third_party_id: str,
        third_party_user_id: str,
        user_context: Dict[str, Any],
    ) -> Union[User, None]:
        params = {
            "thirdPartyId": third_party_id,
            "thirdPartyUserId": third_party_user_id,
        }
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/user"), params
        )
        if "status" in response and response["status"] == "OK":
            return User(
                response["user"]["id"],
                response["user"]["email"],
                response["user"]["timeJoined"],
                ThirdPartyInfo(
                    response["user"]["thirdParty"]["userId"],
                    response["user"]["thirdParty"]["id"],
                ),
            )
        return None

    async def sign_in_up(
        self,
        third_party_id: str,
        third_party_user_id: str,
        email: str,
        user_context: Dict[str, Any],
    ) -> SignInUpOkResult:
        data = {
            "thirdPartyId": third_party_id,
            "thirdPartyUserId": third_party_user_id,
            "email": {"id": email},
        }
        response = await self.querier.send_post_request(
            NormalisedURLPath("/recipe/signinup"), data
        )
        return SignInUpOkResult(
            User(
                response["user"]["id"],
                response["user"]["email"],
                response["user"]["timeJoined"],
                ThirdPartyInfo(
                    response["user"]["thirdParty"]["userId"],
                    response["user"]["thirdParty"]["id"],
                ),
            ),
            response["createdNewUser"],
        )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from pytest import mark
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.asyncio import get_users_newest_first, get_users_oldest_first
from supertokens_python.recipe import emailpassword, session
from supertokens_python.recipe.emailpassword.asyncio import sign_up

from tests.utils import clean_st, reset, setup_st, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@mark.asyncio
async def test_get_users_pagination():
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="https://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[session.init(anti_csrf="VIA_TOKEN"), emailpassword.init()],
    )
    start_st()

    for i in range(5):
        await sign_up(f"dummy{i}@gmail.com", "validpass123")

    # Get all the users (No limit)
    response = await get_users_newest_first()
    assert [user.email for user in response.users] == [
        f"dummy{i}@gmail.com" for i in range(5)
    ][::-1]

    # Get only the oldest user
    response = await get_users_oldest_first(limit=1)
    assert [user.email for user in response.users] == ["dummy0@gmail.com"]

    # Test pagination
    response = await get_users_oldest_first(
        limit=1, pagination_token=response.next_pagination_token
    )
    assert [user.email for user in response.users] == ["dummy1@gmail.com"]



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Dict, List, Union, TypeVar, Callable, Optional

from supertokens_python.recipe.openid.interfaces import (
    GetOpenIdDiscoveryConfigurationResult,
)
from supertokens_python.recipe.session.interfaces import (
    RegenerateAccessTokenOkResult,
    SessionContainer,
    SessionInformationResult,
    SessionClaim,
    SessionClaimValidator,
    SessionDoesNotExistError,
    ClaimsValidationResult,
    JSONObject,
    GetClaimValueOkResult,
)
from supertokens_python.recipe.session.recipe import SessionRecipe
from supertokens_python.types import MaybeAwaitable
from supertokens_python.utils import FRAMEWORKS, resolve, deprecated_warn
from ..utils import get_required_claim_validators
from ...jwt.interfaces import (
    CreateJwtOkResult,
    CreateJwtResultUnsupportedAlgorithm,
    GetJWKSResult,
)

_T = TypeVar("_T")


async def create_new_session(
    request: Any,
    user_id: str,
    access_token_payload: Union[Dict[str, Any], None] = None,
    session_data: Union[Dict[str, Any], None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> SessionContainer:
    if user_context is None:
        user_context = {}
    if session_data is None:
        session_data = {}
    if access_token_payload is None:
        access_token_payload = {}

    claims_added_by_other_recipes = (
        SessionRecipe.get_instance().get_claims_added_by_other_recipes()
    )
    final_access_token_payload = access_token_payload

    for claim in claims_added_by_other_recipes:
        update = await claim.build(user_id, user_context)
        final_access_token_payload = {**final_access_token_payload, **update}

    if not hasattr(request, "wrapper_used") or not request.wrapper_used:
        request = FRAMEWORKS[
            SessionRecipe.get_instance().app_info.framework
        ].wrap_request(request)

    return await SessionRecipe.get_instance().recipe_implementation.create_new_session(
        request,
        user_id,
        final_access_token_payload,
        session_data,
        user_context=user_context,
    )


async def validate_claims_for_session_handle(
    session_handle: str,
    override_global_claim_validators: Optional[
        Callable[
            [
                List[SessionClaimValidator],
                SessionInformationResult,
                Dict[str, Any],
            ],
            MaybeAwaitable[List[SessionClaimValidator]],
        ]
    ] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[SessionDoesNotExistError, ClaimsValidationResult]:
    if user_context is None:
        user_context = {}

    recipe_impl = SessionRecipe.get_instance().recipe_implementation
    session_info = await recipe_impl.get_session_information(
        session_handle, user_context
    )

    if session_info is None:
        return SessionDoesNotExistError()

    claim_validators_added_by_other_recipes = (
        SessionRecipe.get_instance().get_claim_validators_added_by_other_recipes()
    )
    global_claim_validators = await resolve(
        recipe_impl.get_global_claim_validators(
            session_info.user_id,
            claim_validators_added_by_other_recipes,
            user_context,
        )
    )

    if override_global_claim_validators is not None:
        claim_validators = await resolve(
            override_global_claim_validators(
                global_claim_validators, session_info, user_context
            )
        )
    else:
        claim_validators = global_claim_validators

    claim_validation_res = await recipe_impl.validate_claims(
        session_info.user_id,
        session_info.access_token_payload,
        claim_validators,
        user_context,
    )

    if claim_validation_res.access_token_payload_update is not None:
        updated = await recipe_impl.merge_into_access_token_payload(
            session_handle,
            claim_validation_res.access_token_payload_update,
            user_context,
        )
        if not updated:
            return SessionDoesNotExistError()

    return ClaimsValidationResult(claim_validation_res.invalid_claims)


async def validate_claims_in_jwt_payload(
    user_id: str,
    jwt_payload: JSONObject,
    override_global_claim_validators: Optional[
        Callable[
            [
                List[SessionClaimValidator],
                str,
                Dict[str, Any],
            ],
            MaybeAwaitable[List[SessionClaimValidator]],
        ]
    ] = None,
    user_context: Union[None, Dict[str, Any]] = None,
):
    if user_context is None:
        user_context = {}

    recipe_impl = SessionRecipe.get_instance().recipe_implementation

    claim_validators_added_by_other_recipes = (
        SessionRecipe.get_instance().get_claim_validators_added_by_other_recipes()
    )
    global_claim_validators = await resolve(
        recipe_impl.get_global_claim_validators(
            user_id,
            claim_validators_added_by_other_recipes,
            user_context,
        )
    )

    if override_global_claim_validators is not None:
        claim_validators = await resolve(
            override_global_claim_validators(
                global_claim_validators, user_id, user_context
            )
        )
    else:
        claim_validators = global_claim_validators

    return await recipe_impl.validate_claims_in_jwt_payload(
        user_id, jwt_payload, claim_validators, user_context
    )


async def fetch_and_set_claim(
    session_handle: str,
    claim: SessionClaim[Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.fetch_and_set_claim(
        session_handle, claim, user_context
    )


async def get_claim_value(
    session_handle: str,
    claim: SessionClaim[_T],
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[SessionDoesNotExistError, GetClaimValueOkResult[_T]]:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.get_claim_value(
        session_handle, claim, user_context
    )


async def set_claim_value(
    session_handle: str,
    claim: SessionClaim[_T],
    value: _T,
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.set_claim_value(
        session_handle, claim, value, user_context
    )


async def remove_claim(
    session_handle: str,
    claim: SessionClaim[Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.remove_claim(
        session_handle, claim, user_context
    )


async def get_session(
    request: Any,
    anti_csrf_check: Union[bool, None] = None,
    session_required: bool = True,
    override_global_claim_validators: Optional[
        Callable[
            [List[SessionClaimValidator], SessionContainer, Dict[str, Any]],
            MaybeAwaitable[List[SessionClaimValidator]],
        ]
    ] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[SessionContainer, None]:
    if user_context is None:
        user_context = {}
    if not hasattr(request, "wrapper_used") or not request.wrapper_used:
        request = FRAMEWORKS[
            SessionRecipe.get_instance().app_info.framework
        ].wrap_request(request)

    session_recipe_impl = SessionRecipe.get_instance().recipe_implementation
    session = await session_recipe_impl.get_session(
        request,
        anti_csrf_check,
        session_required,
        user_context,
    )

    if session is not None:
        claim_validators = await get_required_claim_validators(
            session, override_global_claim_validators, user_context
        )
        await session.assert_claims(claim_validators, user_context)

    return session


async def refresh_session(
    request: Any, user_context: Union[None, Dict[str, Any]] = None
) -> SessionContainer:
    if user_context is None:
        user_context = {}
    if not hasattr(request, "wrapper_used") or not request.wrapper_used:
        request = FRAMEWORKS[
            SessionRecipe.get_instance().app_info.framework
        ].wrap_request(request)

    return await SessionRecipe.get_instance().recipe_implementation.refresh_session(
        request, user_context
    )


async def revoke_session(
    session_handle: str, user_context: Union[None, Dict[str, Any]] = None
) -> bool:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.revoke_session(
        session_handle, user_context
    )


async def revoke_all_sessions_for_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[str]:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.revoke_all_sessions_for_user(
        user_id, user_context
    )


async def get_all_session_handles_for_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[str]:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.get_all_session_handles_for_user(
        user_id, user_context
    )


async def revoke_multiple_sessions(
    session_handles: List[str], user_context: Union[None, Dict[str, Any]] = None
) -> List[str]:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.revoke_multiple_sessions(
        session_handles, user_context
    )


async def get_session_information(
    session_handle: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[SessionInformationResult, None]:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.get_session_information(
        session_handle, user_context
    )


async def update_session_data(
    session_handle: str,
    new_session_data: Dict[str, Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.update_session_data(
        session_handle, new_session_data, user_context
    )


async def update_access_token_payload(
    session_handle: str,
    new_access_token_payload: Dict[str, Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    if user_context is None:
        user_context = {}

    deprecated_warn(
        "update_access_token_payload is deprecated. Use merge_into_access_token_payload instead"
    )

    return await SessionRecipe.get_instance().recipe_implementation.update_access_token_payload(
        session_handle, new_access_token_payload, user_context
    )


async def merge_into_access_token_payload(
    session_handle: str,
    new_access_token_payload: Dict[str, Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    if user_context is None:
        user_context = {}

    return await SessionRecipe.get_instance().recipe_implementation.merge_into_access_token_payload(
        session_handle, new_access_token_payload, user_context
    )


async def create_jwt(
    payload: Dict[str, Any],
    validity_seconds: Union[None, int] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[CreateJwtOkResult, CreateJwtResultUnsupportedAlgorithm]:
    if user_context is None:
        user_context = {}
    openid_recipe = SessionRecipe.get_instance().openid_recipe

    if openid_recipe is not None:
        return await openid_recipe.recipe_implementation.create_jwt(
            payload, validity_seconds, user_context=user_context
        )

    raise Exception(
        "create_jwt cannot be used without enabling the JWT feature. Please set 'enable: True' for jwt config when initialising the Session recipe"
    )


async def get_jwks(user_context: Union[None, Dict[str, Any]] = None) -> GetJWKSResult:
    if user_context is None:
        user_context = {}
    openid_recipe = SessionRecipe.get_instance().openid_recipe
    if openid_recipe is not None:
        return await openid_recipe.recipe_implementation.get_jwks(user_context)

    raise Exception(
        "get_jwks cannot be used without enabling the JWT feature. Please set 'enable: True' for jwt config when initialising the Session recipe"
    )


async def get_open_id_discovery_configuration(
    user_context: Union[None, Dict[str, Any]] = None
) -> GetOpenIdDiscoveryConfigurationResult:
    if user_context is None:
        user_context = {}
    openid_recipe = SessionRecipe.get_instance().openid_recipe

    if openid_recipe is not None:
        return await openid_recipe.recipe_implementation.get_open_id_discovery_configuration(
            user_context
        )

    raise Exception(
        "get_open_id_discovery_configuration cannot be used without enabling the JWT feature. Please set 'enable: True' for jwt config when initialising the Session recipe"
    )


async def regenerate_access_token(
    access_token: str,
    new_access_token_payload: Union[Dict[str, Any], None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[RegenerateAccessTokenOkResult, None]:
    if user_context is None:
        user_context = {}
    return await SessionRecipe.get_instance().recipe_implementation.regenerate_access_token(
        access_token, new_access_token_payload, user_context
    )



"""
Copyright (c) 2020, VRAI Labs and/or its affiliates. All rights reserved.

This software is licensed under the Apache License, Version 2.0 (the
"License") as published by the Apache Software Foundation.

You may not use this file except in compliance with the License. You may
obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

from _pytest.fixtures import fixture
from fastapi import FastAPI
from pytest import mark
from starlette.requests import Request
from starlette.testclient import TestClient
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import jwt
from supertokens_python.recipe.jwt.interfaces import APIInterface
from supertokens_python.recipe.session.asyncio import create_new_session
from tests.utils import clean_st, reset, setup_st, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    @app.get("/login")
    async def login(request: Request):  # type: ignore
        user_id = "userId"
        await create_new_session(request, user_id, {}, {})
        return {"userId": user_id}

    return TestClient(app)


def apis_override_get_JWKS(param: APIInterface):
    param.disable_jwks_get = True
    return param


@mark.asyncio
async def test_that_default_getJWKS_api_does_not_work_when_disabled(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            jwt.init(override=jwt.OverrideConfig(apis=apis_override_get_JWKS))
        ],
    )
    start_st()

    response = driver_config_client.get(url="/auth/jwt/jwks.json")

    assert response.status_code == 404


@mark.asyncio
async def test_that_default_getJWKS_works_fine(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[jwt.init()],
    )
    start_st()

    response = driver_config_client.get(url="/auth/jwt/jwks.json")

    assert response.status_code == 200
    data = response.json()
    assert len(data["keys"]) > 0



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union

from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.thirdparty import interfaces as ThirdPartyInterfaces
from supertokens_python.recipe.thirdparty.provider import Provider
from supertokens_python.types import APIResponse, GeneralErrorResponse

from ..passwordless import interfaces as PlessInterfaces
from .types import User

# Export re-used classes
ThirdPartyAPIOptions = ThirdPartyInterfaces.APIOptions
PasswordlessAPIOptions = PlessInterfaces.APIOptions

ConsumeCodePostRestartFlowError = PlessInterfaces.ConsumeCodePostRestartFlowError
ConsumeCodePostIncorrectUserInputCodeError = (
    PlessInterfaces.ConsumeCodePostIncorrectUserInputCodeError
)
ConsumeCodePostExpiredUserInputCodeError = (
    PlessInterfaces.ConsumeCodePostExpiredUserInputCodeError
)
ConsumeCodeExpiredUserInputCodeError = (
    PlessInterfaces.ConsumeCodeExpiredUserInputCodeError
)
ConsumeCodeIncorrectUserInputCodeError = (
    PlessInterfaces.ConsumeCodeIncorrectUserInputCodeError
)
ConsumeCodeRestartFlowError = PlessInterfaces.ConsumeCodeRestartFlowError
CreateCodeOkResult = PlessInterfaces.CreateCodeOkResult
CreateCodePostOkResult = PlessInterfaces.CreateCodePostOkResult
CreateNewCodeForDeviceOkResult = PlessInterfaces.CreateNewCodeForDeviceOkResult
CreateNewCodeForDeviceRestartFlowError = (
    PlessInterfaces.CreateNewCodeForDeviceRestartFlowError
)
CreateNewCodeForDeviceUserInputCodeAlreadyUsedError = (
    PlessInterfaces.CreateNewCodeForDeviceUserInputCodeAlreadyUsedError
)
PasswordlessDeleteUserInfoOkResult = PlessInterfaces.DeleteUserInfoOkResult
PasswordlessDeleteUserInfoUnknownUserIdError = (
    PlessInterfaces.DeleteUserInfoUnknownUserIdError
)
DeviceType = PlessInterfaces.DeviceType
PasswordlessEmailExistsGetOkResult = PlessInterfaces.EmailExistsGetOkResult
PasswordlessPhoneNumberExistsGetOkResult = PlessInterfaces.PhoneNumberExistsGetOkResult
ResendCodePostOkResult = PlessInterfaces.ResendCodePostOkResult
ResendCodePostRestartFlowError = PlessInterfaces.ResendCodePostRestartFlowError
RevokeAllCodesOkResult = PlessInterfaces.RevokeAllCodesOkResult
RevokeCodeOkResult = PlessInterfaces.RevokeCodeOkResult
PasswordlessUpdateUserEmailAlreadyExistsError = (
    PlessInterfaces.UpdateUserEmailAlreadyExistsError
)
PasswordlessUpdateUserOkResult = PlessInterfaces.UpdateUserOkResult
PasswordlessUpdateUserPhoneNumberAlreadyExistsError = (
    PlessInterfaces.UpdateUserPhoneNumberAlreadyExistsError
)
PasswordlessUpdateUserUnknownUserIdError = PlessInterfaces.UpdateUserUnknownUserIdError

AuthorisationUrlGetOkResult = ThirdPartyInterfaces.AuthorisationUrlGetOkResult
ThirdPartySignInUpOkResult = ThirdPartyInterfaces.SignInUpOkResult
ThirdPartySignInUpPostNoEmailGivenByProviderResponse = (
    ThirdPartyInterfaces.SignInUpPostNoEmailGivenByProviderResponse
)


class ConsumeCodeOkResult:
    def __init__(self, created_new_user: bool, user: User):
        self.created_new_user = created_new_user
        self.user = user


class RecipeInterface(ABC):
    def __init__(self):
        pass

    @abstractmethod
    async def get_user_by_id(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        pass

    @abstractmethod
    async def get_users_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> List[User]:
        pass

    @abstractmethod
    async def get_user_by_phone_number(
        self, phone_number: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        pass

    @abstractmethod
    async def get_user_by_thirdparty_info(
        self,
        third_party_id: str,
        third_party_user_id: str,
        user_context: Dict[str, Any],
    ) -> Union[User, None]:
        pass

    @abstractmethod
    async def thirdparty_sign_in_up(
        self,
        third_party_id: str,
        third_party_user_id: str,
        email: str,
        user_context: Dict[str, Any],
    ) -> ThirdPartySignInUpOkResult:
        pass

    @abstractmethod
    async def create_code(
        self,
        email: Union[None, str],
        phone_number: Union[None, str],
        user_input_code: Union[None, str],
        user_context: Dict[str, Any],
    ) -> CreateCodeOkResult:
        pass

    @abstractmethod
    async def create_new_code_for_device(
        self,
        device_id: str,
        user_input_code: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        CreateNewCodeForDeviceOkResult,
        CreateNewCodeForDeviceRestartFlowError,
        CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
    ]:
        pass

    @abstractmethod
    async def consume_code(
        self,
        pre_auth_session_id: str,
        user_input_code: Union[str, None],
        device_id: Union[str, None],
        link_code: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        ConsumeCodeOkResult,
        ConsumeCodeIncorrectUserInputCodeError,
        ConsumeCodeExpiredUserInputCodeError,
        ConsumeCodeRestartFlowError,
    ]:
        pass

    @abstractmethod
    async def update_passwordless_user(
        self,
        user_id: str,
        email: Union[str, None],
        phone_number: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        PasswordlessUpdateUserOkResult,
        PasswordlessUpdateUserUnknownUserIdError,
        PasswordlessUpdateUserEmailAlreadyExistsError,
        PasswordlessUpdateUserPhoneNumberAlreadyExistsError,
    ]:
        pass

    @abstractmethod
    async def delete_email_for_passwordless_user(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[
        PasswordlessDeleteUserInfoOkResult, PasswordlessDeleteUserInfoUnknownUserIdError
    ]:
        pass

    @abstractmethod
    async def delete_phone_number_for_user(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[
        PasswordlessDeleteUserInfoOkResult, PasswordlessDeleteUserInfoUnknownUserIdError
    ]:
        pass

    @abstractmethod
    async def revoke_all_codes(
        self,
        email: Union[str, None],
        phone_number: Union[str, None],
        user_context: Dict[str, Any],
    ) -> RevokeAllCodesOkResult:
        pass

    @abstractmethod
    async def revoke_code(
        self, code_id: str, user_context: Dict[str, Any]
    ) -> RevokeCodeOkResult:
        pass

    @abstractmethod
    async def list_codes_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> List[DeviceType]:
        pass

    @abstractmethod
    async def list_codes_by_phone_number(
        self, phone_number: str, user_context: Dict[str, Any]
    ) -> List[DeviceType]:
        pass

    @abstractmethod
    async def list_codes_by_device_id(
        self, device_id: str, user_context: Dict[str, Any]
    ) -> Union[DeviceType, None]:
        pass

    @abstractmethod
    async def list_codes_by_pre_auth_session_id(
        self, pre_auth_session_id: str, user_context: Dict[str, Any]
    ) -> Union[DeviceType, None]:
        pass


class ConsumeCodePostOkResult(APIResponse):
    status: str = "OK"

    def __init__(self, created_new_user: bool, user: User, session: SessionContainer):
        self.created_new_user = created_new_user
        self.user = user
        self.session = session

    def to_json(self):
        user = {"id": self.user.user_id, "time_joined": self.user.time_joined}
        if self.user.email is not None:
            user = {**user, "email": self.user.email}
        if self.user.phone_number is not None:
            user = {**user, "phoneNumber": self.user.phone_number}
        return {
            "status": self.status,
            "createdNewUser": self.created_new_user,
            "user": user,
        }


class ThirdPartySignInUpPostOkResult(APIResponse):
    status: str = "OK"

    def __init__(
        self,
        user: User,
        created_new_user: bool,
        auth_code_response: Dict[str, Any],
        session: SessionContainer,
    ):
        self.user = user
        self.created_new_user = created_new_user
        self.auth_code_response = auth_code_response
        self.session = session

    def to_json(self) -> Dict[str, Any]:
        if self.user.third_party_info is None:
            raise ValueError("Third Party Info cannot be None")

        return {
            "status": self.status,
            "user": {
                "id": self.user.user_id,
                "email": self.user.email,
                "timeJoined": self.user.time_joined,
                "thirdParty": {
                    "id": self.user.third_party_info.id,
                    "userId": self.user.third_party_info.user_id,
                },
            },
            "createdNewUser": self.created_new_user,
        }


class APIInterface(ABC):
    def __init__(self):
        self.disable_thirdparty_sign_in_up_post = False
        self.disable_authorisation_url_get = False
        self.disable_apple_redirect_handler_post = False
        self.disable_create_code_post = False
        self.disable_resend_code_post = False
        self.disable_consume_code_post = False
        self.disable_passwordless_user_email_exists_get = False
        self.disable_passwordless_user_phone_number_exists_get = False

    @abstractmethod
    async def authorisation_url_get(
        self,
        provider: Provider,
        api_options: ThirdPartyAPIOptions,
        user_context: Dict[str, Any],
    ) -> Union[AuthorisationUrlGetOkResult, GeneralErrorResponse]:
        pass

    @abstractmethod
    async def thirdparty_sign_in_up_post(
        self,
        provider: Provider,
        code: str,
        redirect_uri: str,
        client_id: Union[str, None],
        auth_code_response: Union[Dict[str, Any], None],
        api_options: ThirdPartyAPIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        ThirdPartySignInUpPostOkResult,
        ThirdPartySignInUpPostNoEmailGivenByProviderResponse,
        GeneralErrorResponse,
    ]:
        pass

    @abstractmethod
    async def apple_redirect_handler_post(
        self,
        code: str,
        state: str,
        api_options: ThirdPartyAPIOptions,
        user_context: Dict[str, Any],
    ):
        pass

    @abstractmethod
    async def create_code_post(
        self,
        email: Union[str, None],
        phone_number: Union[str, None],
        api_options: PasswordlessAPIOptions,
        user_context: Dict[str, Any],
    ) -> Union[CreateCodePostOkResult, GeneralErrorResponse]:
        pass

    @abstractmethod
    async def resend_code_post(
        self,
        device_id: str,
        pre_auth_session_id: str,
        api_options: PasswordlessAPIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        ResendCodePostOkResult, ResendCodePostRestartFlowError, GeneralErrorResponse
    ]:
        pass

    @abstractmethod
    async def consume_code_post(
        self,
        pre_auth_session_id: str,
        user_input_code: Union[str, None],
        device_id: Union[str, None],
        link_code: Union[str, None],
        api_options: PasswordlessAPIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        ConsumeCodePostOkResult,
        ConsumeCodePostRestartFlowError,
        GeneralErrorResponse,
        ConsumeCodePostIncorrectUserInputCodeError,
        ConsumeCodePostExpiredUserInputCodeError,
    ]:
        pass

    @abstractmethod
    async def passwordless_user_email_exists_get(
        self,
        email: str,
        api_options: PasswordlessAPIOptions,
        user_context: Dict[str, Any],
    ) -> Union[PasswordlessEmailExistsGetOkResult, GeneralErrorResponse]:
        pass

    @abstractmethod
    async def passwordless_user_phone_number_exists_get(
        self,
        phone_number: str,
        api_options: PasswordlessAPIOptions,
        user_context: Dict[str, Any],
    ) -> Union[PasswordlessPhoneNumberExistsGetOkResult, GeneralErrorResponse]:
        pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json
from typing import Any, Dict, Union

from fastapi import Depends, FastAPI
from fastapi.requests import Request
from fastapi.testclient import TestClient
from pytest import fixture, mark
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import emailpassword, session
from supertokens_python.recipe.emailpassword.interfaces import (
    APIInterface as EPAPIInterface,
)
from supertokens_python.recipe.emailpassword.interfaces import APIOptions
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.asyncio import (
    create_new_session,
    get_session,
    refresh_session,
)
from supertokens_python.recipe.session.exceptions import UnauthorisedError
from supertokens_python.recipe.session.framework.fastapi import verify_session
from supertokens_python.recipe.session.interfaces import APIInterface
from supertokens_python.recipe.session.interfaces import APIOptions as SessionAPIOptions
from tests.utils import (
    TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH,
    TEST_DRIVER_CONFIG_COOKIE_DOMAIN,
    TEST_DRIVER_CONFIG_COOKIE_SAME_SITE,
    TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH,
    assert_info_clears_tokens,
    clean_st,
    extract_all_cookies,
    extract_info,
    get_st_init_args,
    reset,
    setup_st,
    start_st,
)


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    @app.get("/login")
    async def login(request: Request):  # type: ignore
        user_id = "userId"
        await create_new_session(request, user_id, {}, {})
        return {"userId": user_id}

    @app.post("/refresh")
    async def custom_refresh(request: Request):  # type: ignore
        await refresh_session(request)
        return {}  # type: ignore

    @app.get("/info")
    async def info_get(request: Request):  # type: ignore
        await get_session(request, True)
        return {}  # type: ignore

    @app.get("/custom/info")
    def custom_info(_):  # type: ignore
        return {}  # type: ignore

    @app.options("/custom/handle")
    def custom_handle_options(_):  # type: ignore
        return {"method": "option"}

    @app.get("/handle")
    async def handle_get(request: Request):  # type: ignore
        session: Union[None, SessionContainer] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        return {"s": session.get_handle()}

    @app.get("/handle-session-optional")
    async def handle_get_optional(session: SessionContainer = Depends(verify_session(session_required=False))):  # type: ignore
        if session is None:
            return {"s": "empty session"}
        return {"s": session.get_handle()}

    @app.post("/logout")
    async def custom_logout(request: Request):  # type: ignore
        session: Union[None, SessionContainer] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        await session.revoke_session()
        return {}  # type: ignore

    @app.post("/create")
    async def _create(request: Request):  # type: ignore
        await create_new_session(request, "userId", {}, {})
        return ""

    @app.post("/create-throw")
    async def _create_throw(request: Request):  # type: ignore
        await create_new_session(request, "userId", {}, {})
        raise UnauthorisedError("unauthorised")

    return TestClient(app)


def apis_override_session(param: APIInterface):
    param.disable_refresh_post = True
    return param


@mark.asyncio
async def test_login_refresh(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                cookie_domain="supertokens.io",
                get_token_transfer_method=lambda _, __, ___: "cookie",
                override=session.InputOverrideConfig(apis=apis_override_session),
            )
        ],
    )
    start_st()

    response_1 = driver_config_client.get("/login")
    cookies_1 = extract_all_cookies(response_1)

    assert response_1.headers.get("anti-csrf") is not None
    assert cookies_1["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_1["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_1["sAccessToken"]["httponly"]
    assert cookies_1["sRefreshToken"]["httponly"]
    assert (
        cookies_1["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_1["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )

    response_3 = driver_config_client.post(
        url="/refresh",
        headers={"anti-csrf": response_1.headers.get("anti-csrf")},
        cookies={
            "sRefreshToken": cookies_1["sRefreshToken"]["value"],
        },
    )
    cookies_3 = extract_all_cookies(response_3)

    assert cookies_3["sAccessToken"]["value"] != cookies_1["sAccessToken"]["value"]
    assert cookies_3["sRefreshToken"]["value"] != cookies_1["sRefreshToken"]["value"]
    assert response_3.headers.get("anti-csrf") is not None
    assert cookies_3["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_3["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_3["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_3["sAccessToken"]["httponly"]
    assert cookies_3["sRefreshToken"]["httponly"]
    assert (
        cookies_3["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_3["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )


@mark.asyncio
async def test_login_logout(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                cookie_domain="supertokens.io",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            )
        ],
    )
    start_st()

    response_1 = driver_config_client.get("/login")
    cookies_1 = extract_all_cookies(response_1)

    assert response_1.headers.get("anti-csrf") is not None
    assert cookies_1["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_1["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_1["sAccessToken"]["httponly"]
    assert cookies_1["sRefreshToken"]["httponly"]
    assert (
        cookies_1["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_1["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert cookies_1["sAccessToken"]["secure"] is None
    assert cookies_1["sRefreshToken"]["secure"] is None

    response_2 = driver_config_client.post(
        url="/logout",
        headers={"anti-csrf": response_1.headers.get("anti-csrf")},
        cookies={
            "sAccessToken": cookies_1["sAccessToken"]["value"],
        },
    )
    cookies_2 = extract_all_cookies(response_2)
    assert response_2.headers.get("anti-csrf") is None
    assert cookies_2["sAccessToken"]["value"] == ""
    assert cookies_2["sRefreshToken"]["value"] == ""
    assert cookies_2["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_2["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_2["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_2["sAccessToken"]["httponly"]
    assert cookies_2["sRefreshToken"]["httponly"]
    assert (
        cookies_2["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_2["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert cookies_2["sAccessToken"]["secure"] is None
    assert cookies_2["sRefreshToken"]["secure"] is None


@mark.asyncio
async def test_login_info(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                cookie_domain="supertokens.io",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            )
        ],
    )
    start_st()

    response_1 = driver_config_client.get("/login")
    cookies_1 = extract_all_cookies(response_1)

    assert response_1.headers.get("anti-csrf") is not None
    assert cookies_1["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_1["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_1["sAccessToken"]["httponly"]
    assert cookies_1["sRefreshToken"]["httponly"]
    assert (
        cookies_1["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_1["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert cookies_1["sAccessToken"]["secure"] is None
    assert cookies_1["sRefreshToken"]["secure"] is None

    response_2 = driver_config_client.get(
        url="/info",
        headers={"anti-csrf": response_1.headers.get("anti-csrf")},
        cookies={
            "sAccessToken": cookies_1["sAccessToken"]["value"],
        },
    )
    cookies_2 = extract_all_cookies(response_2)
    assert not cookies_2


@mark.asyncio
async def test_login_handle(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                cookie_domain="supertokens.io",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            )
        ],
    )
    start_st()

    response_1 = driver_config_client.get("/login")
    cookies_1 = extract_all_cookies(response_1)

    assert response_1.headers.get("anti-csrf") is not None
    assert cookies_1["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_1["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_1["sAccessToken"]["httponly"]
    assert cookies_1["sRefreshToken"]["httponly"]
    assert (
        cookies_1["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_1["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert cookies_1["sAccessToken"]["secure"] is None
    assert cookies_1["sRefreshToken"]["secure"] is None

    response_2 = driver_config_client.get(
        url="/handle",
        headers={"anti-csrf": response_1.headers.get("anti-csrf")},
        cookies={
            "sAccessToken": cookies_1["sAccessToken"]["value"],
        },
    )
    result_dict = json.loads(response_2.content)
    assert "s" in result_dict


@mark.asyncio
async def test_login_refresh_error_handler(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                cookie_domain="supertokens.io",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            )
        ],
    )
    start_st()

    response_1 = driver_config_client.get("/login")
    cookies_1 = extract_all_cookies(response_1)

    assert response_1.headers.get("anti-csrf") is not None
    assert cookies_1["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_1["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_1["sAccessToken"]["httponly"]
    assert cookies_1["sRefreshToken"]["httponly"]
    assert (
        cookies_1["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_1["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert cookies_1["sAccessToken"]["secure"] is None
    assert cookies_1["sRefreshToken"]["secure"] is None

    response_3 = driver_config_client.post(
        url="/refresh",
        headers={"anti-csrf": response_1.headers.get("anti-csrf")},
        cookies={
            # no cookies
        },
    )
    assert response_3.status_code == 401  # not authorized because no refresh tokens


@mark.asyncio
async def test_custom_response(driver_config_client: TestClient):
    def override_email_password_apis(original_implementation: EPAPIInterface):

        original_func = original_implementation.email_exists_get

        async def email_exists_get(
            email: str, api_options: APIOptions, user_context: Dict[str, Any]
        ):
            response_dict = {"custom": True}
            api_options.response.set_status_code(203)
            api_options.response.set_json_content(response_dict)
            return await original_func(email, api_options, user_context)

        original_implementation.email_exists_get = email_exists_get
        return original_implementation

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(
                override=emailpassword.InputOverrideConfig(
                    apis=override_email_password_apis
                )
            )
        ],
    )
    start_st()

    response = driver_config_client.get(
        url="/auth/signup/email/exists?email=test@example.com",
    )

    dict_response = json.loads(response.text)
    assert response.status_code == 203
    assert dict_response["custom"]


@mark.asyncio
async def test_optional_session(driver_config_client: TestClient):

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )
    start_st()

    response = driver_config_client.get(
        url="handle-session-optional",
    )

    dict_response = json.loads(response.text)
    assert response.status_code == 200
    assert dict_response["s"] == "empty session"


@mark.parametrize(
    "fastapi_root_path",
    [
        "/api/v1",
        "/api",
        "api",
        # Don't pass "/" as fastapi_root_path. You'll get unexpected behaviour
        # because api_base_path will be "//auth"
    ],
)
def test_fastapi_root_path(fastapi_root_path: str):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path=f"{fastapi_root_path}/auth",  # It's important to prepend the root path here
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
            emailpassword.init(),
        ],
    )
    start_st()

    # Test with root_path
    app = FastAPI(root_path=fastapi_root_path)
    app.add_middleware(get_middleware())
    test_client = TestClient(app)

    response = test_client.get(
        f"{fastapi_root_path}/auth/signup/email/exists?email=test@example.com"
    )
    assert response.status_code == 200
    assert response.json() == {"status": "OK", "exists": False}

    # The API should migrate (and return 404 here)
    response = test_client.get("/auth/signup/email/exists?email=test@example.com")
    assert response.status_code == 404


@mark.asyncio
@mark.parametrize("token_transfer_method", ["cookie", "header"])
async def test_should_clear_all_response_during_refresh_if_unauthorized(
    driver_config_client: TestClient, token_transfer_method: str
):
    def override_session_apis(oi: APIInterface):
        oi_refresh_post = oi.refresh_post

        async def refresh_post(
            api_options: SessionAPIOptions, user_context: Dict[str, Any]
        ):
            await oi_refresh_post(api_options, user_context)
            raise UnauthorisedError("unauthorized", clear_tokens=True)

        oi.refresh_post = refresh_post
        return oi

    init(
        **get_st_init_args(
            [
                session.init(
                    anti_csrf="VIA_TOKEN",
                    override=session.InputOverrideConfig(apis=override_session_apis),
                )
            ]
        )
    )  # type: ignore
    start_st()

    res = driver_config_client.post(
        "/create", headers={"st-auth-mode": token_transfer_method}
    )
    info = extract_info(res)

    assert info["accessTokenFromAny"] is not None
    assert info["refreshTokenFromAny"] is not None

    headers: Dict[str, Any] = {}
    cookies: Dict[str, Any] = {}

    if token_transfer_method == "header":
        headers.update({"authorization": f"Bearer {info['refreshTokenFromAny']}"})
    else:
        cookies.update(
            {"sRefreshToken": info["refreshTokenFromAny"], "sIdRefreshToken": "asdf"}
        )

    if info["antiCsrf"] is not None:
        headers.update({"anti-csrf": info["antiCsrf"]})

    res = driver_config_client.post(
        "/auth/session/refresh", headers=headers, cookies=cookies
    )
    info = extract_info(res)

    assert res.status_code == 401
    assert_info_clears_tokens(info, token_transfer_method)


@mark.asyncio
@mark.parametrize("token_transfer_method", ["cookie", "header"])
async def test_revoking_session_after_create_new_session_with_throwing_unauthorized_error(
    driver_config_client: TestClient, token_transfer_method: str
):
    init(
        **get_st_init_args(
            [
                session.init(
                    anti_csrf="VIA_TOKEN",
                )
            ]
        )
    )  # type: ignore
    start_st()

    res = driver_config_client.post(
        "/create-throw", headers={"st-auth-mode": token_transfer_method}
    )
    info = extract_info(res)

    assert res.status_code == 401
    assert_info_clears_tokens(info, token_transfer_method)



from typing import Any, Dict, List, Union

from supertokens_python.recipe.userroles.interfaces import (
    AddRoleToUserOkResult,
    CreateNewRoleOrAddPermissionsOkResult,
    DeleteRoleOkResult,
    GetAllRolesOkResult,
    GetPermissionsForRoleOkResult,
    GetRolesForUserOkResult,
    GetRolesThatHavePermissionOkResult,
    GetUsersThatHaveRoleOkResult,
    RemovePermissionsFromRoleOkResult,
    RemoveUserRoleOkResult,
    UnknownRoleError,
)
from supertokens_python.recipe.userroles.recipe import UserRolesRecipe


async def add_role_to_user(
    user_id: str, role: str, user_context: Union[Dict[str, Any], None] = None
) -> Union[AddRoleToUserOkResult, UnknownRoleError]:
    if user_context is None:
        user_context = {}
    return await UserRolesRecipe.get_instance().recipe_implementation.add_role_to_user(
        user_id, role, user_context
    )


async def remove_user_role(
    user_id: str, role: str, user_context: Union[Dict[str, Any], None] = None
) -> Union[RemoveUserRoleOkResult, UnknownRoleError]:
    if user_context is None:
        user_context = {}
    return await UserRolesRecipe.get_instance().recipe_implementation.remove_user_role(
        user_id, role, user_context
    )


async def get_roles_for_user(
    user_id: str, user_context: Union[Dict[str, Any], None] = None
) -> GetRolesForUserOkResult:
    if user_context is None:
        user_context = {}
    return (
        await UserRolesRecipe.get_instance().recipe_implementation.get_roles_for_user(
            user_id, user_context
        )
    )


async def get_users_that_have_role(
    role: str, user_context: Union[Dict[str, Any], None] = None
) -> Union[GetUsersThatHaveRoleOkResult, UnknownRoleError]:
    if user_context is None:
        user_context = {}
    return await UserRolesRecipe.get_instance().recipe_implementation.get_users_that_have_role(
        role, user_context
    )


async def create_new_role_or_add_permissions(
    role: str, permissions: List[str], user_context: Union[Dict[str, Any], None] = None
) -> CreateNewRoleOrAddPermissionsOkResult:
    if user_context is None:
        user_context = {}
    return await UserRolesRecipe.get_instance().recipe_implementation.create_new_role_or_add_permissions(
        role, permissions, user_context
    )


async def get_permissions_for_role(
    role: str, user_context: Union[Dict[str, Any], None] = None
) -> Union[GetPermissionsForRoleOkResult, UnknownRoleError]:
    if user_context is None:
        user_context = {}
    return await UserRolesRecipe.get_instance().recipe_implementation.get_permissions_for_role(
        role, user_context
    )


async def remove_permissions_from_role(
    role: str, permissions: List[str], user_context: Union[Dict[str, Any], None] = None
) -> Union[RemovePermissionsFromRoleOkResult, UnknownRoleError]:
    if user_context is None:
        user_context = {}
    return await UserRolesRecipe.get_instance().recipe_implementation.remove_permissions_from_role(
        role, permissions, user_context
    )


async def get_roles_that_have_permission(
    permission: str, user_context: Union[Dict[str, Any], None] = None
) -> GetRolesThatHavePermissionOkResult:
    if user_context is None:
        user_context = {}
    return await UserRolesRecipe.get_instance().recipe_implementation.get_roles_that_have_permission(
        permission, user_context
    )


async def delete_role(
    role: str, user_context: Union[Dict[str, Any], None] = None
) -> DeleteRoleOkResult:
    if user_context is None:
        user_context = {}
    return await UserRolesRecipe.get_instance().recipe_implementation.delete_role(
        role, user_context
    )


async def get_all_roles(
    user_context: Union[Dict[str, Any], None] = None
) -> GetAllRolesOkResult:
    if user_context is None:
        user_context = {}
    return await UserRolesRecipe.get_instance().recipe_implementation.get_all_roles(
        user_context
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Union

from supertokens_python.ingredients.emaildelivery.types import EmailDeliveryConfig
from supertokens_python.ingredients.smsdelivery.types import SMSDeliveryConfig
from supertokens_python.recipe.passwordless.types import (
    EmailTemplateVars,
    SMSTemplateVars,
)
from typing_extensions import Literal

from . import types, utils
from .emaildelivery import services as emaildelivery_services
from .recipe import PasswordlessRecipe
from .smsdelivery import services as smsdelivery_services

if TYPE_CHECKING:
    from supertokens_python.supertokens import AppInfo

    from ...recipe_module import RecipeModule

InputOverrideConfig = utils.OverrideConfig
ContactEmailOnlyConfig = utils.ContactEmailOnlyConfig
ContactConfig = utils.ContactConfig
PhoneOrEmailInput = utils.PhoneOrEmailInput
CreateAndSendCustomTextMessageParameters = (
    types.CreateAndSendCustomTextMessageParameters
)
CreateAndSendCustomEmailParameters = types.CreateAndSendCustomEmailParameters
ContactPhoneOnlyConfig = utils.ContactPhoneOnlyConfig
ContactEmailOrPhoneConfig = utils.ContactEmailOrPhoneConfig
SMTPService = emaildelivery_services.SMTPService
TwilioService = smsdelivery_services.TwilioService
SuperTokensSMSService = smsdelivery_services.SuperTokensSMSService


def init(
    contact_config: ContactConfig,
    flow_type: Literal[
        "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
    ],
    override: Union[InputOverrideConfig, None] = None,
    get_custom_user_input_code: Union[
        Callable[[Dict[str, Any]], Awaitable[str]], None
    ] = None,
    email_delivery: Union[EmailDeliveryConfig[EmailTemplateVars], None] = None,
    sms_delivery: Union[SMSDeliveryConfig[SMSTemplateVars], None] = None,
) -> Callable[[AppInfo], RecipeModule]:
    return PasswordlessRecipe.init(
        contact_config,
        flow_type,
        override,
        get_custom_user_input_code,
        email_delivery,
        sms_delivery,
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, TypeVar, Union

from twilio.rest import Client  # type: ignore

_T = TypeVar("_T")


class SMSDeliveryInterface(ABC, Generic[_T]):
    @abstractmethod
    async def send_sms(self, template_vars: _T, user_context: Dict[str, Any]) -> None:
        pass


class SMSDeliveryConfig(ABC, Generic[_T]):
    def __init__(
        self,
        service: Union[SMSDeliveryInterface[_T], None] = None,
        override: Union[
            Callable[[SMSDeliveryInterface[_T]], SMSDeliveryInterface[_T]], None
        ] = None,
    ) -> None:
        self.service = service
        self.override = override


class SMSDeliveryConfigWithService(ABC, Generic[_T]):
    def __init__(
        self,
        service: SMSDeliveryInterface[_T],
        override: Union[
            Callable[[SMSDeliveryInterface[_T]], SMSDeliveryInterface[_T]], None
        ] = None,
    ) -> None:
        self.service = service
        self.override = override


class TwilioSettings:
    def __init__(
        self,
        account_sid: str,
        auth_token: str,
        from_: Union[str, None] = None,
        messaging_service_sid: Union[str, None] = None,
        opts: Union[Dict[str, Any], None] = None,
    ) -> None:
        """
        Note: `self.otps` can be used to override values passed to the Twilio Client.
        Read docs from `twilio.rest.Client.__init__` to discover possible args.

        For example, `opts = {"region": "...", "user_agent_extensions": ["..."], }`
        """
        self.account_sid = account_sid
        self.auth_token = auth_token
        self.from_ = from_
        self.messaging_service_sid = messaging_service_sid
        self.opts = opts


class SMSContent:
    def __init__(self, body: str, to_phone: str) -> None:
        self.body = body
        self.to_phone = to_phone


class TwilioServiceInterface(ABC, Generic[_T]):
    def __init__(self, twilio_client: Client) -> None:  # type: ignore
        self.twilio_client = twilio_client  # type: ignore

    @abstractmethod
    async def send_raw_sms(
        self,
        content: SMSContent,
        user_context: Dict[str, Any],
        from_: Union[str, None] = None,
        messaging_service_sid: Union[str, None] = None,
    ) -> None:
        pass

    @abstractmethod
    async def get_content(
        self, template_vars: _T, user_context: Dict[str, Any]
    ) -> SMSContent:
        pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.passwordless.interfaces import APIInterface, APIOptions
from supertokens_python.utils import default_user_context, send_200_response


async def consume_code(api_implementation: APIInterface, api_options: APIOptions):
    if api_implementation.disable_consume_code_post:
        return None

    body = await api_options.request.json()

    if body is None:
        raise_bad_input_exception("Please provide a JSON body")

    user_input_code = None
    device_id = None
    link_code = None

    if "preAuthSessionId" not in body:
        raise_bad_input_exception("Please provide preAuthSessionId")

    if "deviceId" in body or "userInputCode" in body:
        if "linkCode" in body:
            raise_bad_input_exception(
                "Please provide one of (linkCode) or (deviceId+userInputCode) and not both"
            )
        if "deviceId" not in body or "userInputCode" not in body:
            raise_bad_input_exception("Please provide both deviceId and userInputCode")
        device_id = body["deviceId"]
        user_input_code = body["userInputCode"]
    elif "linkCode" in body:
        link_code = body["linkCode"]
    else:
        raise_bad_input_exception(
            "Please provide one of (linkCode) or (deviceId+userInputCode) and not both"
        )

    pre_auth_session_id = body["preAuthSessionId"]
    user_context = default_user_context(api_options.request)

    result = await api_implementation.consume_code_post(
        pre_auth_session_id,
        user_input_code,
        device_id,
        link_code,
        api_options,
        user_context,
    )
    return send_200_response(result.to_json(), api_options.response)



import math
from typing import List, Tuple
from unittest.mock import MagicMock

from pytest import fixture, mark
from pytest_mock import MockerFixture
from supertokens_python.recipe.session.claims import PrimitiveArrayClaim
from supertokens_python.utils import get_timestamp_ms, resolve
from tests.utils import AsyncMock

val = ["a"]
included_item = "a"
excluded_item = "b"

SECONDS = 1_000
MINS = 60 * SECONDS

sync_fetch_value = MagicMock(return_value=val)
async_fetch_value = AsyncMock(return_value=val)

claim_with_inf_max_age: PrimitiveArrayClaim[List[str]] = PrimitiveArrayClaim(
    "key",
    sync_fetch_value,
    math.inf,  # type: ignore
)

pytestmark = mark.asyncio


def teardown_function(_):
    sync_fetch_value.reset_mock()
    async_fetch_value.reset_mock()


@fixture()
def pac_time_patch(mocker: MockerFixture):
    """Patches get_timestamp_ms function of PrimitiveArrayClaim"""
    timestamp = get_timestamp_ms()
    patched_get_timestamp_ms = mocker.patch(
        "supertokens_python.recipe.session.claim_base_classes.primitive_array_claim.get_timestamp_ms"
    )
    patched_get_timestamp_ms.return_value = timestamp

    return patched_get_timestamp_ms, timestamp


@fixture()
def timestamp(pac_time_patch: Tuple[MockerFixture, int]):
    return pac_time_patch[1]


@fixture()
def patch_get_timestamp_ms(pac_time_patch: Tuple[MockerFixture, int]):
    return pac_time_patch[0]


async def test_primitive_claim(timestamp: int):
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    ctx = {}
    res = await claim.build("user_id", ctx)
    assert res == {"key": {"t": timestamp, "v": val}}


async def test_primitive_claim_without_async_fetch_value(timestamp: int):
    claim = PrimitiveArrayClaim("key", async_fetch_value)
    ctx = {}
    res = await claim.build("user_id", ctx)
    assert res == {"key": {"t": timestamp, "v": val}}


async def test_primitive_claim_matching__add_to_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    ctx = {}
    res = await claim.build("user_id", ctx)
    assert res == claim.add_to_payload_({}, val, {})


async def test_primitive_claim_fetch_value_params_correct():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    user_id, ctx = "user_id", {}
    await claim.build(user_id, ctx)
    assert sync_fetch_value.call_count == 1
    assert (user_id, ctx) == sync_fetch_value.call_args_list[0][
        0
    ]  # extra [0] refers to call params


async def test_primitive_claim_fetch_value_none():
    fetch_value_none = MagicMock()
    fetch_value_none.return_value = None

    claim = PrimitiveArrayClaim("key", fetch_value_none)
    user_id, ctx = "user_id", {}
    res = await claim.build(user_id, ctx)
    assert res == {}


# Get value from payload:


async def test_get_value_from_empty_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    assert claim.get_value_from_payload({}) is None


async def test_should_return_value_set_by__add_to_payload_internal():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = claim.add_to_payload_({}, val)
    assert claim.get_value_from_payload(payload) == val


# Get last refetch time:


async def test_get_last_refetch_time_empty_payload():
    claim = PrimitiveArrayClaim("key", async_fetch_value)
    assert claim.get_last_refetch_time({}) is None


async def test_should_return_none_for_empty_payload(timestamp: int):
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")

    assert claim.get_last_refetch_time(payload) == timestamp


# validators.includes


async def test_validators_should_not_validate_empty_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    res = await claim.validators.includes(included_item).validate({}, {})

    assert res.is_valid is False
    assert res.reason == {
        "expectedToInclude": included_item,
        "actualValue": None,
        "message": "value does not exist",
    }


async def test_should_not_validate_mismatching_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")
    res = await claim.validators.includes(excluded_item).validate(payload, {})

    assert res.is_valid is False
    assert res.reason == {
        "expectedToInclude": excluded_item,
        "actualValue": val,
        "message": "wrong value",
    }


async def test_validator_should_validate_matching_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")
    res = await claim.validators.includes(included_item).validate(payload, {})

    assert res.is_valid is True


async def test_should_not_validate_old_values(patch_get_timestamp_ms: MagicMock):
    claim = claim_with_inf_max_age
    payload = await claim.build("user_id")

    # Increase clock time by 1 week
    patch_get_timestamp_ms.return_value += 7 * 24 * 60 * 60 * SECONDS  # type: ignore

    res = await claim.validators.includes(included_item, 600).validate(payload, {})
    assert res.is_valid is False
    assert res.reason == {
        "ageInSeconds": 604800.0,
        "maxAgeInSeconds": 600,
        "message": "expired",
    }


async def test_should_validate_old_values_if_max_age_is_none_and_default_is_inf(
    patch_get_timestamp_ms: MagicMock,
):
    claim = claim_with_inf_max_age
    payload = await claim.build("user_id")

    # Increase clock time by 1 week
    patch_get_timestamp_ms.return_value += 7 * 24 * 60 * 60 * SECONDS  # type: ignore

    res = await claim.validators.includes(included_item).validate(payload, {})
    assert res.is_valid is True


async def test_should_refetch_if_value_not_set():
    claim = claim_with_inf_max_age
    assert await resolve(
        claim.validators.includes(excluded_item, 600).should_refetch({}, {}) is True
    )


async def test_validator_should_not_refetch_if_value_is_set():
    claim = claim_with_inf_max_age
    payload = await claim.build("user_id")
    assert (
        await resolve(
            claim.validators.includes(excluded_item, 600).should_refetch(payload, {})
        )
        is False
    )


async def test_validator_should_refetch_if_value_is_old(
    patch_get_timestamp_ms: MagicMock,
):
    claim = claim_with_inf_max_age
    payload = await claim.build("user_id")

    # Increase clock time by 1 week
    patch_get_timestamp_ms.return_value += 7 * 24 * 60 * 60 * SECONDS  # type: ignore

    assert (
        await resolve(
            claim.validators.includes(excluded_item, 600).should_refetch(payload, {})
        )
        is True
    )


async def test_validator_should_not_refetch_if_max_age_is_none_and_default_is_inf(
    patch_get_timestamp_ms: MagicMock,
):
    claim = claim_with_inf_max_age
    payload = await claim.build("user_id")

    # Increase clock time by 1 week
    patch_get_timestamp_ms.return_value += 7 * 24 * 60 * 60 * SECONDS  # type: ignore

    assert (
        await resolve(
            claim.validators.includes(excluded_item).should_refetch(payload, {})
        )
        is False
    )


async def test_validator_should_validate_values_with_default_max_age(
    patch_get_timestamp_ms: MagicMock,
):
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")

    # Increase clock time by 10 MINS:
    patch_get_timestamp_ms.return_value += 10 * MINS  # type: ignore

    res = await resolve(claim.validators.includes(included_item).validate(payload, {}))
    assert res.is_valid is True


async def test_validator_should_not_refetch_if_max_age_overrides_to_inf(
    patch_get_timestamp_ms: MagicMock,
):
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")

    # Increase clock time by 1 week
    patch_get_timestamp_ms.return_value += 7 * 24 * 60 * 60 * SECONDS  # type: ignore

    assert (
        await resolve(
            claim.validators.includes(
                included_item,
                math.inf,  # type:ignore
            ).should_refetch(payload, {})
        )
        is False
    )


# validator.excludes:


async def test_validator_excludes_should_not_validate_empty_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    res = await claim.validators.excludes(excluded_item).validate({}, {})

    assert res.is_valid is False
    assert res.reason == {
        "expectedToNotInclude": excluded_item,
        "actualValue": None,
        "message": "value does not exist",
    }


async def test_validator_excludes_should_not_validate_mismatching_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")
    res = await claim.validators.excludes(included_item).validate(payload, {})

    assert res.is_valid is False
    assert res.reason == {
        "expectedToNotInclude": included_item,
        "actualValue": val,
        "message": "wrong value",
    }


async def test_validator_excludes_should_validate_matching_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")
    res = await claim.validators.excludes(excluded_item).validate(payload, {})

    assert res.is_valid is True


# validator.includes_all:


async def test_validator_includes_all_should_not_validate_empty_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    res = await claim.validators.includes_all(included_item).validate({}, {})

    assert res.is_valid is False
    assert res.reason == {
        "expectedToInclude": included_item,
        "actualValue": None,
        "message": "value does not exist",
    }


async def test_validator_includes_all_should_not_validate_mismatching_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")
    res = await claim.validators.includes_all(excluded_item).validate(payload, {})

    assert res.is_valid is False
    assert res.reason == {
        "expectedToInclude": excluded_item,
        "actualValue": val,
        "message": "wrong value",
    }


async def test_validator_includes_all_should_validate_matching_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")
    res = await claim.validators.includes_all(included_item).validate(payload, {})

    assert res.is_valid is True


# validator.excludes_all:


async def test_validator_excludes_all_should_not_validate_empty_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    res = await claim.validators.excludes_all(excluded_item).validate({}, {})

    assert res.is_valid is False
    assert res.reason == {
        "expectedToNotInclude": excluded_item,
        "actualValue": None,
        "message": "value does not exist",
    }


async def test_validator_excludes_all_should_not_validate_mismatching_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")
    res = await claim.validators.excludes_all(included_item).validate(payload, {})

    assert res.is_valid is False
    assert res.reason == {
        "expectedToNotInclude": included_item,
        "actualValue": val,
        "message": "wrong value",
    }


async def test_validator_excludes_all_should_validate_matching_payload():
    claim = PrimitiveArrayClaim("key", sync_fetch_value)
    payload = await claim.build("user_id")
    res = await claim.validators.excludes_all(excluded_item).validate(payload, {})

    assert res.is_valid is True


async def test_validator_should_not_validate_older_values_with_5min_default_max_age(
    patch_get_timestamp_ms: MagicMock,
):
    claim = PrimitiveArrayClaim("key", sync_fetch_value, 300)  # 5 mins
    payload = await claim.build("user_id")

    # Increase clock time by 10 MINS:
    patch_get_timestamp_ms.return_value += 10 * MINS  # type: ignore

    res = await resolve(claim.validators.includes(included_item).validate(payload, {}))
    assert res.is_valid is False
    assert res.reason == {
        "ageInSeconds": 600,
        "maxAgeInSeconds": 300,
        "message": "expired",
    }



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing_extensions import Literal

from pytest import mark, skip
from supertokens_python import init
from supertokens_python.interfaces import (
    CreateUserIdMappingOkResult,
)
from supertokens_python.querier import Querier
from supertokens_python.recipe.emailpassword.interfaces import (
    SignUpOkResult,
    ResetPasswordUsingTokenOkResult,
    SignInOkResult,
    CreateResetPasswordOkResult,
)
from supertokens_python.utils import is_version_gte
from tests.utils import clean_st, reset, setup_st, start_st
from .utils import st_config
from supertokens_python.asyncio import (
    create_user_id_mapping,
)


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


pytestmark = mark.asyncio

USER_TYPE = Literal["SUPERTOKENS", "EXTERNAL", "ANY"]


async def ep_get_new_user_id(email: str) -> str:
    from supertokens_python.recipe.emailpassword.asyncio import sign_up

    sign_up_res = await sign_up(email, "password")
    assert isinstance(sign_up_res, SignUpOkResult)

    return sign_up_res.user.user_id


async def ep_get_existing_user_id(user_id: str) -> str:
    from supertokens_python.recipe.emailpassword.asyncio import get_user_by_id

    res = await get_user_by_id(user_id)
    assert res is not None
    return res.user_id


async def ep_get_existing_user_by_email(email: str) -> str:
    from supertokens_python.recipe.emailpassword.asyncio import get_user_by_email

    res = await get_user_by_email(email)
    assert res is not None
    return res.user_id


async def ep_get_existing_user_by_signin(email: str) -> str:
    from supertokens_python.recipe.emailpassword.asyncio import sign_in

    res = await sign_in(email, "password")
    assert isinstance(res, SignInOkResult)
    return res.user.user_id


async def ep_get_existing_user_after_reset_password(user_id: str) -> str:
    new_password = "password"
    from supertokens_python.recipe.emailpassword.asyncio import (
        create_reset_password_token,
        reset_password_using_token,
    )

    result = await create_reset_password_token(user_id)
    assert isinstance(result, CreateResetPasswordOkResult)
    res = await reset_password_using_token(result.token, new_password)
    assert isinstance(res, ResetPasswordUsingTokenOkResult)
    assert res.user_id is not None
    return res.user_id


async def ep_get_existing_user_after_updating_email_and_sign_in(user_id: str) -> str:
    new_email = "bar@example.com"

    from supertokens_python.recipe.emailpassword.asyncio import (
        update_email_or_password,
        sign_in,
    )

    res = await update_email_or_password(user_id, new_email, "password")
    assert isinstance(res, SignUpOkResult)

    res = await sign_in(new_email, "password")
    assert isinstance(res, SignInOkResult)
    return res.user.user_id


@mark.parametrize("use_external_id_info", [(True,), (False,)])
async def test_get_user_id_mapping(use_external_id_info: bool):
    init(**st_config)  # type: ignore
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.15"):
        skip()

    email = "test@example.com"

    supertokens_user_id = await ep_get_new_user_id(email)
    external_user_id = "externalId"
    external_id_info = "externalIdInfo" if use_external_id_info else None

    assert ep_get_existing_user_id(supertokens_user_id) == supertokens_user_id

    # Create user id mapping
    res = await create_user_id_mapping(
        supertokens_user_id, external_user_id, external_id_info
    )
    assert isinstance(res, CreateUserIdMappingOkResult)

    # Now we should get the external user ID instead of ST user ID
    # irrespective of whether we pass ST User ID or External User ID
    assert ep_get_existing_user_id(supertokens_user_id) == external_user_id
    assert ep_get_existing_user_id(external_user_id) == external_user_id

    # Same happens for all the functions
    assert ep_get_existing_user_by_email(email) == external_user_id
    assert ep_get_existing_user_by_signin(email) == external_user_id
    assert (
        ep_get_existing_user_after_reset_password(external_user_id) == external_user_id
    )
    assert (
        ep_get_existing_user_after_updating_email_and_sign_in(external_user_id)
        == external_user_id
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Union

from supertokens_python.ingredients.emaildelivery.types import (
    EmailDeliveryConfig,
    EmailDeliveryConfigWithService,
)
from supertokens_python.ingredients.smsdelivery.types import (
    SMSDeliveryConfig,
    SMSDeliveryConfigWithService,
)
from supertokens_python.recipe.passwordless.types import (
    CreateAndSendCustomEmailParameters,
    PasswordlessLoginSMSTemplateVars,
)
from supertokens_python.utils import deprecated_warn
from typing_extensions import Literal

from .types import CreateAndSendCustomTextMessageParameters

if TYPE_CHECKING:
    from .interfaces import (
        APIInterface,
        RecipeInterface,
        PasswordlessLoginEmailTemplateVars,
    )
    from supertokens_python import AppInfo

from re import fullmatch

from phonenumbers import is_valid_number, parse  # type: ignore
from supertokens_python.recipe.passwordless.emaildelivery.services.backward_compatibility import (
    BackwardCompatibilityService,
)
from supertokens_python.recipe.passwordless.smsdelivery.services.backward_compatibility import (
    BackwardCompatibilityService as SMSBackwardCompatibilityService,
)


async def default_validate_phone_number(value: str):
    try:
        parsed_phone_number: Any = parse(value, None)
        if not is_valid_number(parsed_phone_number):
            return "Phone number is invalid"
    except Exception:
        return "Phone number is invalid"


async def default_validate_email(value: str):
    pattern = r"^(([^<>()\[\]\\.,;:\s@\"]+(\.[^<>()\[\]\\.,;:\s@\"]+)*)|(\".+\"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$"
    if fullmatch(pattern, value) is None:
        return "Email is invalid"


class OverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class ContactConfig(ABC):
    def __init__(self, contact_method: Literal["PHONE", "EMAIL", "EMAIL_OR_PHONE"]):
        self.contact_method = contact_method


class ContactPhoneOnlyConfig(ContactConfig):
    def __init__(
        self,
        create_and_send_custom_text_message: Union[
            Callable[
                [CreateAndSendCustomTextMessageParameters, Dict[str, Any]],
                Awaitable[None],
            ],
            None,
        ] = None,
        validate_phone_number: Union[
            Callable[[str], Awaitable[Union[str, None]]], None
        ] = None,
    ):
        super().__init__("PHONE")

        self.create_and_send_custom_text_message = create_and_send_custom_text_message
        if create_and_send_custom_text_message:
            deprecated_warn(
                "create_and_send_custom_text_message is deprecated. Please use sms delivery config instead"
            )

        if validate_phone_number is None:
            self.validate_phone_number = default_validate_phone_number
        else:
            self.validate_phone_number = validate_phone_number


class ContactEmailOnlyConfig(ContactConfig):
    def __init__(
        self,
        create_and_send_custom_email: Union[
            Callable[
                [CreateAndSendCustomEmailParameters, Dict[str, Any]], Awaitable[None]
            ],
            None,
        ] = None,
        validate_email_address: Union[
            Callable[[str], Awaitable[Union[str, None]]], None
        ] = None,
    ):
        super().__init__("EMAIL")

        self.create_and_send_custom_email = create_and_send_custom_email
        if create_and_send_custom_email is not None:
            deprecated_warn(
                "create_and_send_custom_email is deprecated. Please use email delivery config instead"
            )

        if validate_email_address is None:
            self.validate_email_address = default_validate_email
        else:
            self.validate_email_address = validate_email_address


class ContactEmailOrPhoneConfig(ContactConfig):
    def __init__(
        self,
        create_and_send_custom_email: Union[
            Callable[
                [CreateAndSendCustomEmailParameters, Dict[str, Any]], Awaitable[None]
            ],
            None,
        ] = None,
        create_and_send_custom_text_message: Union[
            Callable[
                [CreateAndSendCustomTextMessageParameters, Dict[str, Any]],
                Awaitable[None],
            ],
            None,
        ] = None,
        validate_email_address: Union[
            Callable[[str], Awaitable[Union[str, None]]], None
        ] = None,
        validate_phone_number: Union[
            Callable[[str], Awaitable[Union[str, None]]], None
        ] = None,
    ):
        super().__init__("EMAIL_OR_PHONE")

        self.create_and_send_custom_email = create_and_send_custom_email
        if create_and_send_custom_email is not None:
            deprecated_warn(
                "create_and_send_custom_email is deprecated. Please use email delivery config instead"
            )

        if validate_email_address is None:
            self.validate_email_address = default_validate_email
        else:
            self.validate_email_address = validate_email_address

        self.create_and_send_custom_text_message = create_and_send_custom_text_message
        if create_and_send_custom_text_message:
            deprecated_warn(
                "create_and_send_custom_text_message is deprecated. Please use sms delivery config instead"
            )

        if validate_phone_number is None:
            self.validate_phone_number = default_validate_phone_number
        else:
            self.validate_phone_number = validate_phone_number


class PhoneOrEmailInput:
    def __init__(self, phone_number: Union[str, None], email: Union[str, None]):
        self.phone_number = phone_number
        self.email = email


class PasswordlessConfig:
    def __init__(
        self,
        contact_config: ContactConfig,
        override: OverrideConfig,
        flow_type: Literal[
            "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
        ],
        get_email_delivery_config: Callable[
            [], EmailDeliveryConfigWithService[PasswordlessLoginEmailTemplateVars]
        ],
        get_sms_delivery_config: Callable[
            [], SMSDeliveryConfigWithService[PasswordlessLoginSMSTemplateVars]
        ],
        get_custom_user_input_code: Union[
            Callable[[Dict[str, Any]], Awaitable[str]], None
        ] = None,
    ):
        self.contact_config = contact_config
        self.override = override
        self.flow_type: Literal[
            "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
        ] = flow_type
        self.get_custom_user_input_code = get_custom_user_input_code
        self.get_email_delivery_config = get_email_delivery_config
        self.get_sms_delivery_config = get_sms_delivery_config


def validate_and_normalise_user_input(
    app_info: AppInfo,
    contact_config: ContactConfig,
    flow_type: Literal[
        "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
    ],
    override: Union[OverrideConfig, None] = None,
    get_custom_user_input_code: Union[
        Callable[[Dict[str, Any]], Awaitable[str]], None
    ] = None,
    email_delivery: Union[
        EmailDeliveryConfig[PasswordlessLoginEmailTemplateVars], None
    ] = None,
    sms_delivery: Union[
        SMSDeliveryConfig[PasswordlessLoginSMSTemplateVars], None
    ] = None,
) -> PasswordlessConfig:

    if override is None:
        override = OverrideConfig()

    def get_email_delivery_config() -> EmailDeliveryConfigWithService[
        PasswordlessLoginEmailTemplateVars
    ]:
        email_service = email_delivery.service if email_delivery is not None else None
        if isinstance(
            contact_config, (ContactEmailOnlyConfig, ContactEmailOrPhoneConfig)
        ):
            create_and_send_custom_email = contact_config.create_and_send_custom_email
        else:
            create_and_send_custom_email = None

        if email_service is None:
            email_service = BackwardCompatibilityService(
                app_info, create_and_send_custom_email
            )

        if email_delivery is not None and email_delivery.override is not None:
            override = email_delivery.override
        else:
            override = None

        return EmailDeliveryConfigWithService(email_service, override=override)

    def get_sms_delivery_config() -> SMSDeliveryConfigWithService[
        PasswordlessLoginSMSTemplateVars
    ]:
        sms_service = sms_delivery.service if sms_delivery is not None else None

        if isinstance(
            contact_config, (ContactPhoneOnlyConfig, ContactEmailOrPhoneConfig)
        ):
            create_and_send_custom_text_message = (
                contact_config.create_and_send_custom_text_message
            )
        else:
            create_and_send_custom_text_message = None

        if sms_service is None:
            sms_service = SMSBackwardCompatibilityService(
                app_info, create_and_send_custom_text_message
            )

        if sms_delivery is not None and sms_delivery.override is not None:
            override = sms_delivery.override
        else:
            override = None

        return SMSDeliveryConfigWithService(sms_service, override=override)

    if not isinstance(contact_config, ContactConfig):  # type: ignore user might not have linter enabled
        raise ValueError("contact_config must be of type ContactConfig")

    if flow_type not in [
        "USER_INPUT_CODE",
        "MAGIC_LINK",
        "USER_INPUT_CODE_AND_MAGIC_LINK",
    ]:
        raise ValueError(
            "flow_type must be one of USER_INPUT_CODE, MAGIC_LINK, USER_INPUT_CODE_AND_MAGIC_LINK"
        )

    if not isinstance(override, OverrideConfig):  # type: ignore user might not have linter enabled
        raise ValueError("override must be of type OverrideConfig")

    return PasswordlessConfig(
        contact_config=contact_config,
        override=OverrideConfig(functions=override.functions, apis=override.apis),
        flow_type=flow_type,
        get_email_delivery_config=get_email_delivery_config,
        get_sms_delivery_config=get_sms_delivery_config,
        get_custom_user_input_code=get_custom_user_input_code,
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json
from math import ceil
from typing import Any, Dict, Optional

from supertokens_python.framework.response import BaseResponse
from supertokens_python.utils import get_timestamp_ms


class FastApiResponse(BaseResponse):
    from fastapi import Response

    def __init__(self, response: Response):
        super().__init__({})
        self.response = response
        self.original = response
        self.parser_checked = False
        self.response_sent = False
        self.status_set = False

    def set_html_content(self, content: str):
        if not self.response_sent:
            body = bytes(content, "utf-8")
            self.set_header("Content-Length", str(len(body)))
            self.set_header("Content-Type", "text/html")
            self.response.body = body
            self.response_sent = True

    def set_cookie(
        self,
        key: str,
        value: str,
        expires: int,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: str = "lax",
    ):
        # Note: For FastAPI response object, the expires value
        # doesn't mean the absolute time in ms, but the duration in seconds
        # So we need to convert our absolute expiry time (ms) to a duration (seconds)

        # we do ceil because if we do floor, we tests may fail where the access
        # token lifetime is set to 1 second
        self.response.set_cookie(
            key=key,
            value=value,  # Note: Unlike other frameworks, FastAPI wraps the value in quotes in Set-Cookie header
            expires=ceil((expires - get_timestamp_ms()) / 1000),
            path=path,
            domain=domain,  # type: ignore # starlette didn't set domain as optional type but their default value is None anyways
            secure=secure,
            httponly=httponly,
            samesite=samesite,
        )

    def set_header(self, key: str, value: str):
        self.response.headers[key] = value

    def get_header(self, key: str) -> Optional[str]:
        return self.response.headers.get(key, None)

    def remove_header(self, key: str):
        del self.response.headers[key]

    def set_status_code(self, status_code: int):
        if not self.status_set:
            self.response.status_code = status_code
            self.status_code = status_code
            self.status_set = True

    def set_json_content(self, content: Dict[str, Any]):
        if not self.response_sent:
            body = json.dumps(
                content,
                ensure_ascii=False,
                allow_nan=False,
                indent=None,
                separators=(",", ":"),
            ).encode("utf-8")
            self.set_header("Content-Type", "application/json; charset=utf-8")
            self.set_header("Content-Length", str(len(body)))
            self.response.body = body
            self.response_sent = True



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from os import environ
from typing import TYPE_CHECKING, List, Union

from supertokens_python.querier import Querier
from supertokens_python.recipe.jwt import JWTRecipe

from .api.implementation import APIImplementation
from .api.open_id_discovery_configuration_get import open_id_discovery_configuration_get
from .constants import GET_DISCOVERY_CONFIG_URL
from .exceptions import SuperTokensOpenIdError
from .interfaces import APIOptions
from .recipe_implementation import RecipeImplementation
from .utils import InputOverrideConfig, validate_and_normalise_user_input

if TYPE_CHECKING:
    from supertokens_python.framework.request import BaseRequest
    from supertokens_python.framework.response import BaseResponse
    from supertokens_python.supertokens import AppInfo

from supertokens_python.exceptions import SuperTokensError, raise_general_exception
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.recipe_module import APIHandled, RecipeModule


class OpenIdRecipe(RecipeModule):
    recipe_id = "openid"
    __instance = None

    def __init__(
        self,
        recipe_id: str,
        app_info: AppInfo,
        jwt_validity_seconds: Union[int, None] = None,
        issuer: Union[str, None] = None,
        override: Union[InputOverrideConfig, None] = None,
    ):
        super().__init__(recipe_id, app_info)
        self.config = validate_and_normalise_user_input(app_info, issuer, override)
        jwt_feature = None
        if override is not None:
            jwt_feature = override.jwt_feature
        self.jwt_recipe = JWTRecipe(
            recipe_id, app_info, jwt_validity_seconds, jwt_feature
        )

        recipe_implementation = RecipeImplementation(
            Querier.get_instance(recipe_id),
            self.config,
            app_info,
            self.jwt_recipe.recipe_implementation,
        )
        self.recipe_implementation = (
            recipe_implementation
            if self.config.override.functions is None
            else self.config.override.functions(recipe_implementation)
        )
        api_implementation = APIImplementation()
        self.api_implementation = (
            api_implementation
            if self.config.override.apis is None
            else self.config.override.apis(api_implementation)
        )

    def get_apis_handled(self) -> List[APIHandled]:
        return [
            APIHandled(
                method="get",
                path_without_api_base_path=NormalisedURLPath(GET_DISCOVERY_CONFIG_URL),
                request_id=GET_DISCOVERY_CONFIG_URL,
                disabled=self.api_implementation.disable_open_id_discovery_configuration_get,
            )
        ] + self.jwt_recipe.get_apis_handled()

    async def handle_api_request(
        self,
        request_id: str,
        request: BaseRequest,
        path: NormalisedURLPath,
        method: str,
        response: BaseResponse,
    ):
        options = APIOptions(
            request,
            response,
            self.get_recipe_id(),
            self.config,
            self.recipe_implementation,
        )

        if request_id == GET_DISCOVERY_CONFIG_URL:
            return await open_id_discovery_configuration_get(
                self.api_implementation, options
            )
        return await self.jwt_recipe.handle_api_request(
            request_id, request, path, method, response
        )

    async def handle_error(
        self, request: BaseRequest, err: SuperTokensError, response: BaseResponse
    ):
        if isinstance(err, SuperTokensOpenIdError):
            raise err
        return await self.jwt_recipe.handle_error(request, err, response)

    def get_all_cors_headers(self) -> List[str]:
        return self.jwt_recipe.get_all_cors_headers()

    def is_error_from_this_recipe_based_on_instance(self, err: Exception) -> bool:
        return isinstance(err, SuperTokensError) and (
            isinstance(err, SuperTokensOpenIdError)
            or self.jwt_recipe.is_error_from_this_recipe_based_on_instance(err)
        )

    @staticmethod
    def init(
        jwt_validity_seconds: Union[int, None] = None,
        issuer: Union[str, None] = None,
        override: Union[InputOverrideConfig, None] = None,
    ):
        def func(app_info: AppInfo):
            if OpenIdRecipe.__instance is None:
                OpenIdRecipe.__instance = OpenIdRecipe(
                    OpenIdRecipe.recipe_id,
                    app_info,
                    jwt_validity_seconds,
                    issuer,
                    override,
                )
                return OpenIdRecipe.__instance
            raise_general_exception(
                "OpenId recipe has already been initialised. Please check your code for bugs."
            )

        return func

    @staticmethod
    def get_instance() -> OpenIdRecipe:
        if OpenIdRecipe.__instance is not None:
            return OpenIdRecipe.__instance
        raise_general_exception(
            "Initialisation not done. Did you forget to call the SuperTokens.init function?"
        )

    @staticmethod
    def reset():
        if ("SUPERTOKENS_ENV" not in environ) or (
            environ["SUPERTOKENS_ENV"] != "testing"
        ):
            raise_general_exception("calling testing function in non testing env")
        OpenIdRecipe.__instance = None



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from supertokens_python import SupertokensConfig, InputAppInfo
from supertokens_python.recipe import emailpassword, session, usermetadata

st_config = {
    "supertokens_config": SupertokensConfig("http://localhost:3567"),
    "app_info": InputAppInfo(
        app_name="SuperTokens Demo",
        api_domain="https://api.supertokens.io",
        website_domain="supertokens.io",
    ),
    "framework": "fastapi",
    "recipe_list": [
        emailpassword.init(),
        usermetadata.init(),
        session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
    ],
}



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Optional

from typing_extensions import Literal


class UnknownSupertokensUserIDError:
    pass


class CreateUserIdMappingOkResult:
    pass


class UserIdMappingAlreadyExistsError:
    def __init__(
        self, does_super_tokens_user_id_exist: bool, does_external_user_id_exist: str
    ):
        self.does_super_tokens_user_id_exist = does_super_tokens_user_id_exist
        self.does_external_user_id_exist = does_external_user_id_exist


UserIDTypes = Literal["SUPERTOKENS", "EXTERNAL", "ANY"]


class GetUserIdMappingOkResult:
    def __init__(
        self,
        supertokens_user_id: str,
        external_user_id: str,
        external_user_info: Optional[str] = None,
    ):
        self.supertokens_user_id = supertokens_user_id
        self.external_user_id = external_user_id
        self.external_user_info = external_user_info


class UnknownMappingError:
    pass


class DeleteUserIdMappingOkResult:
    def __init__(self, did_mapping_exist: bool):
        self.did_mapping_exist = did_mapping_exist


class UpdateOrDeleteUserIdMappingInfoOkResult:
    pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any, Dict

from supertokens_python.constants import DASHBOARD_VERSION
from supertokens_python.framework import BaseRequest
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.querier import Querier

from .interfaces import RecipeInterface
from .utils import DashboardConfig, validate_api_key


class RecipeImplementation(RecipeInterface):
    async def get_dashboard_bundle_location(self, user_context: Dict[str, Any]) -> str:
        return f"https://cdn.jsdelivr.net/gh/supertokens/dashboard@v{DASHBOARD_VERSION}/build/"

    async def should_allow_access(
        self,
        request: BaseRequest,
        config: DashboardConfig,
        user_context: Dict[str, Any],
    ) -> bool:
        if config.auth_mode == "email-password":
            auth_header_value = request.get_header("authorization")

            if not auth_header_value:
                return False

            auth_header_value = auth_header_value.split()[1]
            session_verification_response = (
                await Querier.get_instance().send_post_request(
                    NormalisedURLPath("/recipe/dashboard/session/verify"),
                    {"sessionId": auth_header_value},
                )
            )
            return (
                "status" in session_verification_response
                and session_verification_response["status"] == "OK"
            )
        return validate_api_key(request, config)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json
from typing import Union

from fastapi import FastAPI
from fastapi.requests import Request
from fastapi.testclient import TestClient
from pytest import fixture, mark
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import emailpassword, session
from supertokens_python.recipe.emailpassword.interfaces import APIInterface
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.asyncio import (
    create_new_session,
    get_session,
    refresh_session,
)
from tests.utils import clean_st, reset, setup_st, sign_up_request, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    @app.get("/login")
    async def login(request: Request):  # type: ignore
        user_id = "userId"
        await create_new_session(request, user_id, {}, {})
        return {"userId": user_id}

    @app.post("/refresh")
    async def custom_refresh(request: Request):  # type: ignore
        await refresh_session(request)
        return {}  # type: ignore

    @app.get("/info")
    async def info_get(request: Request):  # type: ignore
        await get_session(request, True)
        return {}  # type: ignore

    @app.get("/custom/info")
    def custom_info(_):  # type: ignore
        return {}  # type: ignore

    @app.options("/custom/handle")
    def custom_handle_options(_):  # type: ignore
        return {"method": "option"}

    @app.get("/handle")
    async def handle_get(request: Request):  # type: ignore
        session: Union[None, SessionContainer] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        return {"s": session.get_handle()}

    @app.post("/logout")
    async def custom_logout(request: Request):  # type: ignore
        session: Union[None, SessionContainer] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        await session.revoke_session()
        return {}  # type: ignore

    return TestClient(app)


def apis_override_email_password(param: APIInterface):
    param.disable_email_exists_get = True
    return param


@mark.asyncio
async def test_good_input_email_exists(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(anti_csrf="VIA_TOKEN", cookie_domain="supertokens.io"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validPass123"
    )

    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.get(
        url="/auth/signup/email/exists", params={"email": "random@gmail.com"}
    )
    assert response_2.status_code == 200


@mark.asyncio
async def test_good_input_email_does_not_exists(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(anti_csrf="VIA_TOKEN", cookie_domain="supertokens.io"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = driver_config_client.get(
        url="/auth/signup/email/exists", params={"email": "random@gmail.com"}
    )

    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    assert dict_response["exists"] is False


@mark.asyncio
async def test_that_if_disabling_api_the_default_email_exists_api_does_not_work(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(anti_csrf="VIA_TOKEN", cookie_domain="supertokens.io"),
            emailpassword.init(
                override=emailpassword.InputOverrideConfig(
                    apis=apis_override_email_password
                )
            ),
        ],
    )
    start_st()

    response_1 = driver_config_client.get(
        url="/auth/signup/email/exists", params={"email": "random@gmail.com"}
    )

    assert response_1.status_code == 404


@mark.asyncio
async def test_email_exists_a_syntactically_invalid_email(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(anti_csrf="VIA_TOKEN", cookie_domain="supertokens.io"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validPass123"
    )

    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.get(
        url="/auth/signup/email/exists", params={"email": "randomgmail.com"}
    )
    assert response_2.status_code == 200
    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"
    assert dict_response["exists"] is False


@mark.asyncio
async def test_sending_an_unnormalised_email_and_you_get_exists_is_true(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(anti_csrf="VIA_TOKEN", cookie_domain="supertokens.io"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validPass123"
    )

    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.get(
        url="/auth/signup/email/exists", params={"email": "rAndOM@gmAiL.COm"}
    )
    assert response_2.status_code == 200
    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"
    assert dict_response["exists"] is True


@mark.asyncio
async def test_bad_input_do_not_pass_email(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(anti_csrf="VIA_TOKEN", cookie_domain="supertokens.io"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validPass123"
    )

    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.get(url="/auth/signup/email/exists", params={})
    assert response_2.status_code == 400
    assert "Please provide the email as a GET param" in response_2.text


@mark.asyncio
async def test_passing_an_array_instead_of_a_string_in_the_email(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(anti_csrf="VIA_TOKEN", cookie_domain="supertokens.io"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validPass123"
    )

    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.get(
        url="/auth/signup/email/exists",
        params={"email": ["rAndOM@gmAiL.COm", "x@g.com"]},
    )
    assert response_2.status_code == 200



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Union
from urllib.parse import parse_qsl

from supertokens_python.framework.request import BaseRequest

if TYPE_CHECKING:
    from supertokens_python.recipe.session.interfaces import SessionContainer


class DjangoRequest(BaseRequest):
    from django.http import HttpRequest

    def __init__(self, request: HttpRequest):
        super().__init__()
        self.request = request

    def get_query_param(
        self, key: str, default: Union[str, None] = None
    ) -> Union[str, None]:
        return self.request.GET.get(key, default)

    async def json(self) -> Union[Any, None]:
        try:
            body = json.loads(self.request.body)
            return body
        except Exception:
            return {}

    def method(self) -> str:
        if self.request.method is None:
            raise Exception("Should never come here")
        return self.request.method

    def get_cookie(self, key: str) -> Union[str, None]:
        return self.request.COOKIES.get(key)

    def get_header(self, key: str) -> Union[None, str]:
        key = key.replace("-", "_")
        key = "HTTP_" + key
        return self.request.META.get(key.upper())

    def get_session(self) -> Union[SessionContainer, None]:
        return self.request.supertokens  # type: ignore

    def set_session(self, session: SessionContainer):
        self.request.supertokens = session  # type: ignore

    def set_session_as_none(self):
        self.request.supertokens = None  # type: ignore

    def get_path(self) -> str:
        return self.request.path

    async def form_data(self):
        return dict(parse_qsl(self.request.body.decode("utf-8")))



from unittest.mock import MagicMock

from supertokens_python import init
from supertokens_python.framework import BaseRequest
from supertokens_python.recipe import session
from supertokens_python.recipe.session.asyncio import create_new_session
from tests.utils import (
    setup_function,
    teardown_function,
    start_st,
    min_api_version,
    st_init_common_args,
)
from .utils import (
    NoneClaim,
    get_st_init_args,
    session_functions_override_with_claim,
    TrueClaim,
)

_ = setup_function  # type:ignore
_ = teardown_function  # type:ignore


@min_api_version("2.13")
async def test_create_access_token_payload_with_session_claims(timestamp: int):
    init(**get_st_init_args(TrueClaim))  # type:ignore
    start_st()

    dummy_req: BaseRequest = MagicMock()
    s = await create_new_session(dummy_req, "someId")

    payload = s.get_access_token_payload()
    assert payload == {"st-true": {"v": True, "t": timestamp}}


@min_api_version("2.13")
async def test_should_create_access_token_payload_with_session_claims_with_an_none_value():
    init(**get_st_init_args(NoneClaim))  # type:ignore
    start_st()

    dummy_req: BaseRequest = MagicMock()
    s = await create_new_session(dummy_req, "someId")

    payload = s.get_access_token_payload()
    assert payload == {}


@min_api_version("2.13")
async def test_should_merge_claims_and_passed_access_token_payload_obj(timestamp: int):
    new_st_init = {
        **st_init_common_args,
        "recipe_list": [
            session.init(
                override=session.InputOverrideConfig(
                    functions=session_functions_override_with_claim(
                        TrueClaim, {"user-custom-claim": "foo"}
                    ),
                )
            ),
        ],
    }
    init(**new_st_init)  # type:ignore
    start_st()

    dummy_req: BaseRequest = MagicMock()
    s = await create_new_session(dummy_req, "someId")

    payload = s.get_access_token_payload()
    assert payload == {
        "st-true": {"v": True, "t": timestamp},
        "user-custom-claim": "foo",
    }



"""
Copyright (c) 2020, VRAI Labs and/or its affiliates. All rights reserved.

This software is licensed under the Apache License, Version 2.0 (the
"License") as published by the Apache Software Foundation.

You may not use this file except in compliance with the License. You may
obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

from pytest import mark
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.recipe import jwt
from supertokens_python.recipe.jwt import JWTRecipe
from tests.utils import clean_st, reset, setup_st, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@mark.asyncio
async def test_that_the_default_config_sets_values_correctly_for_JWT_recipe():
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[jwt.init()],
    )
    start_st()

    jwt_recipe = JWTRecipe.get_instance()
    assert jwt_recipe.config.jwt_validity_seconds == 3153600000


@mark.asyncio
async def test_that_the_config_sets_values_correctly_for_JWT_recipe_when_jwt_validity_is_set():
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[jwt.init(jwt_validity_seconds=24 * 60 * 60)],  # 24 hours
    )
    start_st()

    jwt_recipe = JWTRecipe.get_instance()
    assert jwt_recipe.config.jwt_validity_seconds == 24 * 60 * 60



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Union

from supertokens_python.logger import log_debug_message
from supertokens_python.recipe.emailpassword.constants import (
    FORM_FIELD_EMAIL_ID,
    FORM_FIELD_PASSWORD_ID,
)
from supertokens_python.recipe.emailpassword.interfaces import (
    APIInterface,
    CreateResetPasswordWrongUserIdError,
    EmailExistsGetOkResult,
    GeneratePasswordResetTokenPostOkResult,
    PasswordResetPostInvalidTokenResponse,
    PasswordResetPostOkResult,
    ResetPasswordUsingTokenInvalidTokenError,
    SignInPostOkResult,
    SignInPostWrongCredentialsError,
    SignInWrongCredentialsError,
    SignUpEmailAlreadyExistsError,
    SignUpPostEmailAlreadyExistsError,
    SignUpPostOkResult,
)
from supertokens_python.recipe.emailpassword.types import (
    FormField,
    PasswordResetEmailTemplateVars,
    PasswordResetEmailTemplateVarsUser,
)
from supertokens_python.recipe.session.asyncio import create_new_session
from supertokens_python.utils import find_first_occurrence_in_list

if TYPE_CHECKING:
    from supertokens_python.recipe.emailpassword.interfaces import APIOptions

from supertokens_python.types import GeneralErrorResponse


class APIImplementation(APIInterface):
    async def email_exists_get(
        self, email: str, api_options: APIOptions, user_context: Dict[str, Any]
    ) -> Union[EmailExistsGetOkResult, GeneralErrorResponse]:
        user = await api_options.recipe_implementation.get_user_by_email(
            email, user_context
        )
        return EmailExistsGetOkResult(user is not None)

    async def generate_password_reset_token_post(
        self,
        form_fields: List[FormField],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[GeneratePasswordResetTokenPostOkResult, GeneralErrorResponse]:
        emailFormField = find_first_occurrence_in_list(
            lambda x: x.id == FORM_FIELD_EMAIL_ID, form_fields
        )
        if emailFormField is None:
            raise Exception("Should never come here")
        email = emailFormField.value

        user = await api_options.recipe_implementation.get_user_by_email(
            email, user_context
        )

        if user is None:
            return GeneratePasswordResetTokenPostOkResult()

        token_result = (
            await api_options.recipe_implementation.create_reset_password_token(
                user.user_id, user_context
            )
        )

        if isinstance(token_result, CreateResetPasswordWrongUserIdError):
            log_debug_message(
                "Password reset email not sent, unknown user id: %s", user.user_id
            )
            return GeneratePasswordResetTokenPostOkResult()

        token = token_result.token
        password_reset_link = (
            api_options.app_info.website_domain.get_as_string_dangerous()
            + api_options.app_info.website_base_path.get_as_string_dangerous()
            + "/reset-password?token="
            + token
            + "&rid="
            + api_options.recipe_id
        )

        log_debug_message("Sending password reset email to %s", email)
        send_email_input = PasswordResetEmailTemplateVars(
            user=PasswordResetEmailTemplateVarsUser(user.user_id, user.email),
            password_reset_link=password_reset_link,
        )
        await api_options.email_delivery.ingredient_interface_impl.send_email(
            send_email_input, user_context
        )

        return GeneratePasswordResetTokenPostOkResult()

    async def password_reset_post(
        self,
        form_fields: List[FormField],
        token: str,
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        PasswordResetPostOkResult,
        PasswordResetPostInvalidTokenResponse,
        GeneralErrorResponse,
    ]:
        new_password_for_field = find_first_occurrence_in_list(
            lambda x: x.id == FORM_FIELD_PASSWORD_ID, form_fields
        )
        if new_password_for_field is None:
            raise Exception("Should never come here")
        new_password = new_password_for_field.value

        result = await api_options.recipe_implementation.reset_password_using_token(
            token, new_password, user_context
        )

        if isinstance(result, ResetPasswordUsingTokenInvalidTokenError):
            return PasswordResetPostInvalidTokenResponse()

        return PasswordResetPostOkResult(result.user_id)

    async def sign_in_post(
        self,
        form_fields: List[FormField],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        SignInPostOkResult, SignInPostWrongCredentialsError, GeneralErrorResponse
    ]:
        password_form_field = find_first_occurrence_in_list(
            lambda x: x.id == FORM_FIELD_PASSWORD_ID, form_fields
        )
        if password_form_field is None:
            raise Exception("Should never come here")
        password = password_form_field.value

        email_form_field = find_first_occurrence_in_list(
            lambda x: x.id == FORM_FIELD_EMAIL_ID, form_fields
        )
        if email_form_field is None:
            raise Exception("Should never come here")
        email = email_form_field.value

        result = await api_options.recipe_implementation.sign_in(
            email, password, user_context
        )

        if isinstance(result, SignInWrongCredentialsError):
            return SignInPostWrongCredentialsError()

        user = result.user
        session = await create_new_session(
            api_options.request,
            user.user_id,
            user_context=user_context,
        )
        return SignInPostOkResult(user, session)

    async def sign_up_post(
        self,
        form_fields: List[FormField],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        SignUpPostOkResult, SignUpPostEmailAlreadyExistsError, GeneralErrorResponse
    ]:
        password_form_field = find_first_occurrence_in_list(
            lambda x: x.id == FORM_FIELD_PASSWORD_ID, form_fields
        )
        if password_form_field is None:
            raise Exception("Should never come here")
        password = password_form_field.value

        email_form_field = find_first_occurrence_in_list(
            lambda x: x.id == FORM_FIELD_EMAIL_ID, form_fields
        )
        if email_form_field is None:
            raise Exception("Should never come here")
        email = email_form_field.value

        result = await api_options.recipe_implementation.sign_up(
            email, password, user_context
        )

        if isinstance(result, SignUpEmailAlreadyExistsError):
            return SignUpPostEmailAlreadyExistsError()

        user = result.user
        session = await create_new_session(
            api_options.request,
            user.user_id,
            user_context=user_context,
        )
        return SignUpPostOkResult(user, session)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from supertokens_python.recipe.emailpassword.interfaces import (
        APIOptions,
        APIInterface,
    )

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.utils import default_user_context, send_200_response

from .utils import validate_form_fields_or_throw_error


async def handle_generate_password_reset_token_api(
    api_implementation: APIInterface, api_options: APIOptions
):
    if api_implementation.disable_generate_password_reset_token_post:
        return None
    body = await api_options.request.json()
    if body is None:
        raise_bad_input_exception("Please provide a JSON body")
    form_fields_raw: Any = body["formFields"] if "formFields" in body else []
    form_fields = await validate_form_fields_or_throw_error(
        api_options.config.reset_password_using_token_feature.form_fields_for_generate_token_form,
        form_fields_raw,
    )
    user_context = default_user_context(api_options.request)

    response = await api_implementation.generate_password_reset_token_post(
        form_fields, api_options, user_context
    )
    return send_200_response(response.to_json(), api_options.response)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from os import environ
from typing import TYPE_CHECKING, Any, Dict, List, Union, Callable, Optional

from supertokens_python.framework.response import BaseResponse
from typing_extensions import Literal

from supertokens_python.utils import default_user_context

from .api import handle_refresh_api, handle_signout_api
from .cookie_and_header import get_cors_allowed_headers
from .exceptions import (
    SuperTokensSessionError,
    TokenTheftError,
    UnauthorisedError,
    InvalidClaimsError,
)
from ...types import MaybeAwaitable

if TYPE_CHECKING:
    from supertokens_python.framework import BaseRequest
    from supertokens_python.supertokens import AppInfo

from supertokens_python.exceptions import SuperTokensError, raise_general_exception
from supertokens_python.logger import log_debug_message
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.querier import Querier
from supertokens_python.recipe.openid.recipe import OpenIdRecipe
from supertokens_python.recipe.session.with_jwt import (
    get_recipe_implementation_with_jwt,
)
from supertokens_python.recipe_module import APIHandled, RecipeModule

from .api.implementation import APIImplementation
from .constants import SESSION_REFRESH, SIGNOUT
from .interfaces import (
    APIInterface,
    APIOptions,
    RecipeInterface,
    SessionClaim,
    SessionClaimValidator,
    SessionContainer,
)
from .recipe_implementation import (
    RecipeImplementation,
)
from .utils import (
    InputErrorHandlers,
    InputOverrideConfig,
    TokenTransferMethod,
    JWTConfig,
    validate_and_normalise_user_input,
)


class SessionRecipe(RecipeModule):
    recipe_id = "session"
    __instance = None

    def __init__(
        self,
        recipe_id: str,
        app_info: AppInfo,
        cookie_domain: Union[str, None] = None,
        cookie_secure: Union[bool, None] = None,
        cookie_same_site: Union[Literal["lax", "none", "strict"], None] = None,
        session_expired_status_code: Union[int, None] = None,
        anti_csrf: Union[
            Literal["VIA_TOKEN", "VIA_CUSTOM_HEADER", "NONE"], None
        ] = None,
        get_token_transfer_method: Union[
            Callable[
                [BaseRequest, bool, Dict[str, Any]],
                Union[TokenTransferMethod, Literal["any"]],
            ],
            None,
        ] = None,
        error_handlers: Union[InputErrorHandlers, None] = None,
        override: Union[InputOverrideConfig, None] = None,
        jwt: Union[JWTConfig, None] = None,
        invalid_claim_status_code: Union[int, None] = None,
    ):
        super().__init__(recipe_id, app_info)
        self.openid_recipe: Union[None, OpenIdRecipe] = None
        self.config = validate_and_normalise_user_input(
            app_info,
            cookie_domain,
            cookie_secure,
            cookie_same_site,
            session_expired_status_code,
            anti_csrf,
            get_token_transfer_method,
            error_handlers,
            override,
            jwt,
            invalid_claim_status_code,
        )
        log_debug_message("session init: anti_csrf: %s", self.config.anti_csrf)
        if self.config.cookie_domain is not None:
            log_debug_message(
                "session init: cookie_domain: %s", self.config.cookie_domain
            )
        else:
            log_debug_message("session init: cookie_domain: None")
        log_debug_message(
            "session init: cookie_same_site: %s", self.config.cookie_same_site
        )
        log_debug_message(
            "session init: cookie_secure: %s", str(self.config.cookie_secure)
        )
        log_debug_message(
            "session init: refresh_token_path: %s ",
            self.config.refresh_token_path.get_as_string_dangerous(),
        )
        log_debug_message(
            "session init: session_expired_status_code: %s",
            str(self.config.session_expired_status_code),
        )

        if self.config.jwt.enable:
            openid_feature_override = None
            if override is not None:
                openid_feature_override = override.openid_feature
            self.openid_recipe = OpenIdRecipe(
                recipe_id,
                app_info,
                None,
                self.config.jwt.issuer,
                openid_feature_override,
            )
            recipe_implementation = RecipeImplementation(
                Querier.get_instance(recipe_id), self.config, self.app_info
            )
            recipe_implementation = get_recipe_implementation_with_jwt(
                recipe_implementation,
                self.config,
                self.openid_recipe.recipe_implementation,
            )
        else:
            recipe_implementation = RecipeImplementation(
                Querier.get_instance(recipe_id), self.config, self.app_info
            )
        self.recipe_implementation: RecipeInterface = (
            recipe_implementation
            if self.config.override.functions is None
            else self.config.override.functions(recipe_implementation)
        )
        api_implementation = APIImplementation()
        self.api_implementation: APIInterface = (
            api_implementation
            if self.config.override.apis is None
            else self.config.override.apis(api_implementation)
        )

        self.claims_added_by_other_recipes: List[SessionClaim[Any]] = []
        self.claim_validators_added_by_other_recipes: List[SessionClaimValidator] = []

    def is_error_from_this_recipe_based_on_instance(self, err: Exception) -> bool:
        return isinstance(err, SuperTokensError) and (
            isinstance(err, SuperTokensSessionError)
            or (
                self.openid_recipe is not None
                and self.openid_recipe.is_error_from_this_recipe_based_on_instance(err)
            )
        )

    def get_apis_handled(self) -> List[APIHandled]:
        apis_handled = [
            APIHandled(
                NormalisedURLPath(SESSION_REFRESH),
                "post",
                SESSION_REFRESH,
                self.api_implementation.disable_refresh_post,
            ),
            APIHandled(
                NormalisedURLPath(SIGNOUT),
                "post",
                SIGNOUT,
                self.api_implementation.disable_signout_post,
            ),
        ]
        if self.openid_recipe is not None:
            apis_handled = apis_handled + self.openid_recipe.get_apis_handled()

        return apis_handled

    async def handle_api_request(
        self,
        request_id: str,
        request: BaseRequest,
        path: NormalisedURLPath,
        method: str,
        response: BaseResponse,
    ) -> Union[BaseResponse, None]:
        if request_id == SESSION_REFRESH:
            return await handle_refresh_api(
                self.api_implementation,
                APIOptions(
                    request,
                    response,
                    self.recipe_id,
                    self.config,
                    self.recipe_implementation,
                ),
            )
        if request_id == SIGNOUT:
            return await handle_signout_api(
                self.api_implementation,
                APIOptions(
                    request,
                    response,
                    self.recipe_id,
                    self.config,
                    self.recipe_implementation,
                ),
            )
        if self.openid_recipe is not None:
            return await self.openid_recipe.handle_api_request(
                request_id, request, path, method, response
            )
        return None

    async def handle_error(
        self, request: BaseRequest, err: SuperTokensError, response: BaseResponse
    ) -> BaseResponse:
        if isinstance(err, SuperTokensSessionError):
            for mutator in err.response_mutators:
                mutator(response)

        if isinstance(err, UnauthorisedError):
            log_debug_message("errorHandler: returning UNAUTHORISED")
            return await self.config.error_handlers.on_unauthorised(
                self, err.clear_tokens, request, str(err), response
            )
        if isinstance(err, TokenTheftError):
            log_debug_message("errorHandler: returning TOKEN_THEFT_DETECTED")
            return await self.config.error_handlers.on_token_theft_detected(
                self, request, err.session_handle, err.user_id, response
            )
        if isinstance(err, InvalidClaimsError):
            log_debug_message("errorHandler: returning INVALID_CLAIMS")
            return await self.config.error_handlers.on_invalid_claim(
                self, request, err.payload, response
            )

        log_debug_message("errorHandler: returning TRY_REFRESH_TOKEN")
        return await self.config.error_handlers.on_try_refresh_token(
            request, str(err), response
        )

    def get_all_cors_headers(self) -> List[str]:
        cors_headers = get_cors_allowed_headers()
        if self.openid_recipe is not None:
            cors_headers = cors_headers + self.openid_recipe.get_all_cors_headers()

        return cors_headers

    @staticmethod
    def init(
        cookie_domain: Union[str, None] = None,
        cookie_secure: Union[bool, None] = None,
        cookie_same_site: Union[Literal["lax", "none", "strict"], None] = None,
        session_expired_status_code: Union[int, None] = None,
        anti_csrf: Union[
            Literal["VIA_TOKEN", "VIA_CUSTOM_HEADER", "NONE"], None
        ] = None,
        get_token_transfer_method: Union[
            Callable[
                [BaseRequest, bool, Dict[str, Any]],
                Union[TokenTransferMethod, Literal["any"]],
            ],
            None,
        ] = None,
        error_handlers: Union[InputErrorHandlers, None] = None,
        override: Union[InputOverrideConfig, None] = None,
        jwt: Union[JWTConfig, None] = None,
        invalid_claim_status_code: Union[int, None] = None,
    ):
        def func(app_info: AppInfo):
            if SessionRecipe.__instance is None:
                SessionRecipe.__instance = SessionRecipe(
                    SessionRecipe.recipe_id,
                    app_info,
                    cookie_domain,
                    cookie_secure,
                    cookie_same_site,
                    session_expired_status_code,
                    anti_csrf,
                    get_token_transfer_method,
                    error_handlers,
                    override,
                    jwt,
                    invalid_claim_status_code,
                )
                return SessionRecipe.__instance
            raise_general_exception(
                "Session recipe has already been initialised. Please check your code for bugs."
            )

        return func

    @staticmethod
    def get_instance() -> SessionRecipe:
        if SessionRecipe.__instance is not None:
            return SessionRecipe.__instance
        raise_general_exception(
            "Initialisation not done. Did you forget to call the SuperTokens.init function?"
        )

    @staticmethod
    def reset():
        if ("SUPERTOKENS_ENV" not in environ) or (
            environ["SUPERTOKENS_ENV"] != "testing"
        ):
            raise_general_exception("calling testing function in non testing env")
        SessionRecipe.__instance = None

    def add_claim_from_other_recipe(self, claim: SessionClaim[Any]):
        # We are throwing here (and not in addClaimValidatorFromOtherRecipe) because if multiple
        # claims are added with the same key they will overwrite each other. Validators will all run
        # and work as expected even if they are added multiple times.
        if claim.key in [c.key for c in self.claims_added_by_other_recipes]:
            raise Exception("Claim added by multiple recipes")

        self.claims_added_by_other_recipes.append(claim)

    def get_claims_added_by_other_recipes(self) -> List[SessionClaim[Any]]:
        return self.claims_added_by_other_recipes

    def add_claim_validator_from_other_recipe(
        self, claim_validator: SessionClaimValidator
    ):
        self.claim_validators_added_by_other_recipes.append(claim_validator)

    def get_claim_validators_added_by_other_recipes(
        self,
    ) -> List[SessionClaimValidator]:
        return self.claim_validators_added_by_other_recipes

    async def verify_session(
        self,
        request: BaseRequest,
        anti_csrf_check: Union[bool, None],
        session_required: bool,
        override_global_claim_validators: Optional[
            Callable[
                [List[SessionClaimValidator], SessionContainer, Dict[str, Any]],
                MaybeAwaitable[List[SessionClaimValidator]],
            ]
        ],
        user_context: Dict[str, Any],
    ):
        _ = user_context

        return await self.api_implementation.verify_session(
            APIOptions(
                request,
                None,
                self.recipe_id,
                self.config,
                self.recipe_implementation,
            ),
            anti_csrf_check,
            session_required,
            override_global_claim_validators,
            user_context=default_user_context(request),
        )
